import ast
import json
import os
import re
import sys
import tempfile
import time
from functools import wraps
from typing import Any, Callable, Iterable
from colorama import Style
from prompt_toolkit.enums import EditingMode
import hupper
from prompt_toolkit import PromptSession
import pytest

from gentest.util import PROJECT_DIR
from gentest.validate.errors import GenException
from gentest.harness.database import expand_short_path
from gentest.cli.collect_failures import FailInfo, FailureCollectorPlugin
from gentest.cli.collect_tests import collect_tests, pytest_node_fn, pytest_node_name
from gentest.cli.repro import Shareable, encode_shareable, export_ir, export_pyrel, get_example, get_fn_file, reproduce, reproduce_blob



class State:
    should_die = False
    session_start: float
    runs = 0
    repro_target: str|None = None
    export_ir: str|None = None
    export_pyrel: str|None = None

    def __init__(self, state_file_path: str):
        self.state_file_path = state_file_path

    def get(self, key: str, default: Any):
        return vars(self).get(key, default)

    def load(self):
        with open(self.state_file_path, 'r') as state_file:
            state = json.load(state_file)
            for key, value in state.items():
                setattr(self, key, value)

    def dump(self):
        state = {k: v for k, v in vars(self).items() if not callable(v)}
        with open(self.state_file_path, 'w') as state_file:
            json.dump(state, state_file)

    def has_exports(self):
        return bool(self.export_pyrel or self.export_ir)

    def do_exports(self):
        if self.repro_target and self.has_exports():
            ex = get_example(self.repro_target)
            if self.export_ir:
                export_ir(self.repro_target, self.export_ir, ex)
            if self.export_pyrel:
                export_pyrel(self.repro_target, self.export_pyrel, ex)

    def get_target(self) -> str:
        if not self.repro_target:
            raise Exception("No target failure for operation!")

        return self.repro_target

    def __repr__(self):
        kwargs = ", ".join(f"{k}={repr(v)}" for k, v in vars(self).items() if not callable(v))
        return f"State({kwargs})"

failure_collector = FailureCollectorPlugin()

def run_tests(state: State):
    sys.stdout.write('\r\033[K')
    sys.stdout.flush()
    print()
    print(f" RUN {state.runs} ".center(80, "#"))
    repro_target = state.get("repro_target", None)
    if repro_target:
        reproduce(repro_target)
        state.do_exports()
    else:
        pytest.main(['-s'], plugins=[failure_collector])

    state.runs += 1
    state.dump()

def get_failure(ix: int):
    failures = failure_collector.get_failures()
    if ix > len(failures) or ix < 1:
        raise ArgumentError(f"'{ix}' is not a valid failure index from the last run ({len(failures)} failure(s)).")

    return failures[ix - 1]

def failure(by: str):
    try:
        ix = int(by)
        if ix == 0:
            raise ArgumentError("Failures are 1-indexed for convenience. Try `target 1` instead.")
        return get_failure(ix)

    except ValueError:
        if not re.match(r'^[a-zA-Z0-9]+/[a-zA-Z0-9]+$', by):
            raise ArgumentError(f"{by} is not a valid failure index or path")

        key_hash, value_hash = by.split("/")
        for failure in failure_collector.get_failures():
            err = failure.error
            if isinstance(err, GenException):
                if err.key_hash and err.key_hash.startswith(key_hash) and err.value_hash().startswith(value_hash):
                    return failure

        raise ArgumentError(f"{by} does not correspond to an existing failure index or path. Try `list` to see valid options.")

def failure_path(by: str):
    try:
        ix = int(by)
        if ix == 0:
            raise ArgumentError("Failures are 1-indexed for convenience. Try `target 1` instead.")
        fail_info = get_failure(ix)
        err = fail_info.error
        if not isinstance(err, GenException):
            print(f"Provided ix '{ix}' is not a gen exception, so most repro functionality will not work with it.")
            print(str(err))
            print(type(err))
            return

        return err.hash_path()
    except ValueError:
        if not re.match(r'^[a-zA-Z0-9]+/[a-zA-Z0-9]+$', by):
            raise ArgumentError(f"{by} is not a valid failure index or path")

        try:
            by = expand_short_path(by, relative=True)
        except Exception as err:
            raise ArgumentError(f"{by} is not a valid failure index or path")
        return by

    

ArgParser = Callable[[str], Iterable[Any]]
Action = Callable[[str], Any]

class ArgumentError(Exception):
    ops: list[str]

    def __init__(self, message: str):
        super().__init__(message)
        self.ops = []
        self.message = message

    def prefix_op(self, op: str):
        self.ops.insert(0, op)
        return self

    def __str__(self):
        if self.ops:
            return f"[{' '.join(self.ops)}] {self.message}"
        return self.message


def arg_parser(parser: ArgParser):
    def decorator(func: Callable):
        @wraps(func)
        def wrapper(user_input: str):
            return func(*parser(user_input))

        help_str = getattr(parser, "__help_str", None)
        if help_str is None:
            print(f"WARN: arg parser {parser.__name__} does not have help text configured. Use @help.")
        else:
            setattr(wrapper, "__arg_help_str", help_str)
        return wrapper
    return decorator

def help(help_text: str):
    def decorator(func):
        setattr(func, "__help_str", help_text)
        return func

    return decorator

@help('')
def no_args(_: str):
    return ()

class CommandSet:
    actions: list[tuple[Iterable[str], Action]]
    call_path: list[str]

    __name__: str = "subcommand"
    __help_str: str|None = None
    __arg_help_str: str|None = None
    _ = __help_str, __arg_help_str # shut up pyright

    def __init__(self, help_str: str|None = None, name: str|None = None):
        self.actions = []
        self.call_path = []
        self.add(('h', 'help', '?'), self.print_help_action)
        self.__help_str = help_str
        self.__arg_help_str = self.get_help_str()
        setattr(self, '__name__', name)

    def __call__(self, args: str):
        try:
            self.dispatch(*self.dispatch_parser(args))
        except TypeError as err:
            if str(err).startswith("CommandSet.dispatch()"):
                print(f"{Style.BRIGHT}Missing subcommand. Valid options are:{Style.NORMAL}\n")
                self.print_help()
            else:
                raise

    def dispatch(self, op: str, args: str = ''):
        for key, action in self.actions:
            if op in key:
                try:
                    if isinstance(action, CommandSet):
                        action.call_path = self.call_path + [op]
                    action(args)
                except ArgumentError as err:
                    raise err.prefix_op(op)

                break
        else:
            print(f"{Style.BRIGHT}Unrecognized subcommand '{op}'. Valid options are:{Style.NORMAL}\n")
            self.print_help()

    def dispatch_parser(self, raw: str):
        return raw.split(maxsplit=1)

    def get_help_str(self):
        ops = '|'.join(op for (ops, _) in self.actions for op in ops)
        return f"<op: {ops}> [args...]"

    def add(self, op: str|Iterable[str], action: Action):
        if isinstance(op, str):
            op = (op,)

        self.actions.insert(-1, (op, action))
        # self.get_help_str()
        return self


    @help("Print this help text.")
    def print_help_action(self, _: str):
        return self.print_help()

    def print_help(self, indent = 0):
        pad = "    " * indent
        max_op_len = 0
        max_arg_len = 0
        op_texts = []
        arg_texts = []
        desc_texts = []
        for key, action in self.actions:
            op_text = '|'.join(key)
            if len(op_text) > max_op_len:
                max_op_len = len(op_text)
            arg_text = getattr(action, '__arg_help_str', '')
            if len(arg_text) > max_arg_len:
                max_arg_len = len(arg_text)

            op_texts.append(op_text)
            arg_texts.append(arg_text)
            desc_texts.append(getattr(action, '__help_str', action.__name__))

        for (op, arg, desc) in zip(op_texts, arg_texts, desc_texts):
            print(f"{pad}{op:<{max_op_len}}  {arg:<{max_arg_len}} -- {desc}")


def simple_kwarg_parser(kwarg_parsers: dict[str, Callable[[str], Any]]) -> ArgParser:
    help_str = ' '.join(f"<{key}: {parser.__name__}>" for key, parser in kwarg_parsers.items())
    @help(help_str)
    def parse(raw: str):
        args = raw.split()
        res = []
        expected_len = len(kwarg_parsers.keys())
        actual_len = len(args)
        if(expected_len != actual_len):
            print(f"Invalid arguments, expecting: {help_str}")
            raise ArgumentError(f"expected {expected_len} arguments, got {actual_len}.")
        for key, arg in zip(kwarg_parsers.keys(), args):
            res.append(kwarg_parsers[key](arg))

        return res
    return parse

def build_cmd_tree(state: State):
    @arg_parser(simple_kwarg_parser({"out": str}))
    @help("Export the pickled IR from the targeted failure to OUT")
    def export_ir_action(out: str):
        export_ir(state.get_target(), out)
        state.export_ir = out
        state.dump()

    @arg_parser(simple_kwarg_parser({"out": str}))
    @help("Export the emitted PyRel from the targeted failure to OUT")
    def export_pyrel_action(out: str):
        export_pyrel(state.get_target(), out)
        state.export_pyrel = out
        state.dump()

    @arg_parser(no_args)
    @help("Encode a binary blob that will allow others to reproduce the targeted failure")
    def export_blob_action():
        print(repr(encode_shareable(state.get_target())))

    @arg_parser(no_args)
    @help("Clear all active exports")
    def export_clear_action():
        state.export_ir = None
        state.export_pyrel = None
        print("Clearing active exports.")
        state.dump()

    export_cmd = (
        CommandSet("Export the example generated in the targeted failure.")
        .add(('ir', 'i'), export_ir_action)
        .add(('pyrel', 'p'), export_pyrel_action)
        .add(('blob', 'b'), export_blob_action)
        .add(('clear', 'c'), export_clear_action)
    )

    @arg_parser(simple_kwarg_parser({"target": failure_path}))
    @help("Focus the specified failure, only re-running it on change.")
    def target_action(short_path: str):
        state.repro_target = short_path
        state.dump()
        run_tests(state)

    @arg_parser(no_args)
    @help("Clear the targeted failure and any running operations like export that depend on it.")
    def clear_action():
        was_targeted = state.repro_target is not None
        state.repro_target = None
        state.export_ir = None
        state.export_pyrel = None
        state.dump()

        if was_targeted:
            run_tests(state)

    @arg_parser(no_args)
    @help("Print a short list of the accumulated failures from the last run.")
    def list_action():
        for ix, failure in enumerate(failure_collector.get_failures(), start=1):
            match failure.error:
                case GenException():
                    prefix = f"{ix}. {failure.error.short}"
                    print(f"{prefix:<80}[{failure.error.hash_path()}]")
                case _:
                    msg = str(failure.error)
                    print(f"{ix}. {msg[:75]}{'…' if len(msg) > 75 else ''}")

    @arg_parser(simple_kwarg_parser({"target": failure}))
    @help("Show details for the specified failure.")
    def show_action(failure: FailInfo):
        target = pytest_node_fn(failure.node)
        target = getattr(target, "inner", target)
        if hasattr(target, "hypothesis"):
            target = target.hypothesis.inner_test

        file = target.__code__.co_filename
        name = pytest_node_name(failure.node)
        print(f"file     {os.path.relpath(file, PROJECT_DIR)}")
        print(f"test     {name}")
        if isinstance(failure.error, GenException):
            print(f"failpath {failure.error.hash_path()}")
        print(failure.error)

    @help("<shareable: tuple>")
    def shareable_blob_parser(raw: str):
        try:
            shareable = ast.literal_eval(raw)
            if isinstance(shareable, tuple) and len(shareable) == 3:
                return shareable,
        except (ValueError, SyntaxError):
            pass

        raise ArgumentError("Argument should be in the format `('test/file/path', 'test_name', b'reproblob')`. Produce a shareable blob with `e b`")

    @arg_parser(shareable_blob_parser)
    def import_action(shareable: Shareable):
        test_path, test_name, blob = shareable
        tests = collect_tests()
        for test in tests:
            print(pytest_node_name(test), get_fn_file(pytest_node_fn(test).inner))
            if pytest_node_name(test) != test_name:
                continue
            fn = pytest_node_fn(test)
            if get_fn_file(getattr(fn, "inner", fn)) == test_path:
                break
        else:
            raise Exception(f"Unable to find specified test {test_name} in {test_path}.")

        reproduce_blob(fn, blob)

    @arg_parser(no_args)
    @help("Quit gentest.")
    def quit_action():
        print("Goodbye!")
        state.should_die = True
        state.dump()

    return (
        CommandSet("Operate on the failing tests.")
        .add(('target', 't'), target_action)
        .add(('clear', 'c'), clear_action)
        .add(('export', 'e'), export_cmd)
        .add(('list', 'l'), list_action)
        .add(('show', 's'), show_action)
        .add(('import', 'i'), import_action)
        .add(('quit', 'q'), quit_action)
    )

def run_repl(state: State):
    session = PromptSession()
    cmd = build_cmd_tree(state)
    while True:
        try:
            # Prompt for user input
            user_input: str = session.prompt('> ', editing_mode=EditingMode.EMACS)
            if not user_input.strip():
                cmd.print_help()
            else:
                cmd(user_input)
                if state.should_die:
                    break
        except ArgumentError as err:
            print(str(err))
        except (EOFError, KeyboardInterrupt):
            print("Goodbye!")
            break


# Main function to start test watching and interactive CLI
def run(state_file_path: str|None = None):
    owns_state_file = False
    if state_file_path is None:
        _, state_file_path = tempfile.mkstemp(prefix='gentest_cli_state_', suffix='.json')
        with open(state_file_path,  'w') as state_file:
            json.dump({'session_start': time.time()}, state_file)
            owns_state_file = True

    try:
        reloader: Any = hupper.start_reloader("gentest.cli.watch.run", worker_args=(state_file_path,), verbose=0)

        state = State(state_file_path)
        state.load()
        run_tests(state)
        run_repl(state)

        # Violently kill the parent process by feeding it an unexpected message, since hupper lacks this functionality :/
        reloader.pipe.send("exit")
        sys.exit(0)

    except RuntimeError as err:
        if not str(err).endswith("('received unknown control signal', 'exit')"):
            raise
    finally:
        try:
            if owns_state_file:
                os.remove(state_file_path)
        except FileNotFoundError:
            pass

if __name__ == "__main__":
    run()



from ..dsl import alias, rel, Vars
from . import graphs, aggregates, strings, math, dates

__all__ = [
    "aggregates",
    "alias",
    "dates",
    "graphs",
    "math",
    "rel",
    "strings",
    "Vars"
]


from typing import cast
from ..metamodel import Action, Builtins, Task, ActionType, Behavior, Var
import json
from ..rel import Emitter
import rich

rel_emitter = Emitter()

#--------------------------------------------------
# Mehcanism
#--------------------------------------------------

class Mechanism():
    def __init__(self, task:Task) -> None:
        self.stack = []
        self.sequence()
        self.task = task
        self.interpret(task)

    #--------------------------------------------------
    # IDs
    #--------------------------------------------------

    current_id = 0

    @staticmethod
    def get_next_id():
        Mechanism.current_id += 1
        return Mechanism.current_id

    #--------------------------------------------------
    # Containers
    #--------------------------------------------------

    def sequence(self):
        self.stack.append({"id": self.get_next_id(), "type": "sequence", "items": []})

    def union(self, result):
        self.stack.append({"id": self.get_next_id(), "type": "union", "items": [], "result": self._var(result)})

    def choice(self):
        self.stack.append({"id": self.get_next_id(), "type": "choice", "items": []})

    def pop(self):
        cur = self.stack.pop()
        self.stack[-1]["items"].append(cur)

    #--------------------------------------------------
    # Actions
    #--------------------------------------------------

    def _append(self, item):
        self.stack[-1]["items"].append(item)

    def _var(self, var:Var):
        if var.value is not None:
            return str(var)
        else:
            return {"id": var.id, "name": var.name}

    def get(self, entity, types, props):
        props = {key: self._var(value) for key, value in props.items()}
        self._append({"id": self.get_next_id(), "type": "get", "entity": self._var(entity), "types": types, "props": props})

    def filter(self, op, args):
        self._append({"id": self.get_next_id(), "type": "filter", "op":op, "args": [self._var(arg) for arg in args]})

    def compute(self, op, args, ret):
        self._append({"id": self.get_next_id(), "type": "compute", "op":op, "args": [self._var(arg) for arg in args], "ret": self._var(ret)})

    def effect(self, op, entity, types, props):
        props = {key: self._var(value) for key, value in props.items()}
        self._append({"id": self.get_next_id(), "type": "effect", "op": op, "entity": self._var(entity), "types": types, "props": props})

    def return_(self, vars):
        self._append({"id": self.get_next_id(), "type": "return", "values":[self._var(v) for v in vars]})

    def install(self, item):
        self._append({"id": self.get_next_id(), "type": "install", "item":self._var(item)})

    def aggregate(self, op, args, group, ret):
        self._append({"id": self.get_next_id(), "type": "aggregate", "op":op, "args": [self._var(arg) for arg in args], "group": [self._var(g) for g in group], "ret": self._var(ret)})

    def quantify(self, quantifier, group):
        self._append({"id": self.get_next_id(), "type": "quantify", "quantifier": quantifier, "group": [self._var(g) for g in group]})

    def to_dict(self):
        return self.stack[-1]

    def to_json(self):
        return json.dumps(self.to_dict(), indent=4)

    def debug(self, item = None, indent = 0):
        pad = "    " * indent
        if not item:
            item = self.stack[-1]

        def var(v):
            if isinstance(v, dict):
                return f"<{v['name']} ({v['id']})>"
            return v

        if "items" in item:
            res = f" -> {var(item['result'])}" if "result" in item else ""
            rich.print(f"{pad}{item['type']}{res}")
            for child in item['items']:
                if item is not None:
                    self.debug(child, indent + 1)

        elif item['type'] == "get" or item['type'] == "effect":
            rich.print(f"{pad}{item['type']} {var(item['entity'])} {item['types']} {{{', '.join(f'{k}={var(v)}' for k, v in item['props'].items())}}}")
        elif item['type'] == "filter":
            rich.print(f"{pad}{item['type']} {item['op']} [{', '.join(var(v) for v in item['args'])}]")
        elif item['type'] == "compute":
            rich.print(f"{pad}{item['type']} {item['op']} [{', '.join(var(v) for v in item['args'])}] -> {var(item['ret'])}")
        elif item['type'] == "return":
            rich.print(f"{pad}{item['type']} [{', '.join(var(v) for v in item['values'])}]")
        elif item['type'] == "aggregate":
            rich.print(f"{pad}{item['type']} {item['op']} [{', '.join(var(v) for v in item['args'])}] per [{', '.join(var(v) for v in item['group'])}] -> {item['ret']}")
        elif item['type'] == "quantify":
            rich.print(f"{pad}{item['type']} {item['quantifier']} per [{', '.join(var(v) for v in item['group'])}]")
        else:
            rich.print(f"{pad}{item['type']} ???")

    #--------------------------------------------------
    # Interpret
    #--------------------------------------------------

    def find_min(self, item:Action):
        if item.is_subtask_call():
            return min(getattr(item.entity.value, "id", item.id), *[self.find_min(subtask) for subtask in cast(Task, item.entity.value).items])
        elif item.entity.isa(Builtins.Quantifier):
            return cast(Task, [*item.bindings.values()][1].value).id
        return item.id

    def interpret(self, task):
        # hack to get around the way stratification currently works, in our mechanistic
        # diagram we want to see the effects last in queries. This will change once we
        # have order dependence in the v2 compiler
        items = task.items[:]
        # Subtask calls get added during the dsl stack collapse and so end up with a much
        # later id. Their "place" in the original query is determined by the id of the
        # subtask rather than the call
        items.sort(key=self.find_min)
        for item in items:
            if item.action == ActionType.Get:
                for type_ in item.types:
                    self.get(item.entity, [type_.name], {})
                for key, value in item.bindings.items():
                    self.get(item.entity, [], {key.name: value})
            elif item.action == ActionType.Call:
                func = item.entity
                params = list(item.bindings.values())
                if item.is_subtask_call():
                    sub_task = func.value
                    if sub_task.behavior == Behavior.Query or sub_task.behavior == Behavior.Sequence:
                        self.sequence()
                        self.interpret(sub_task)
                        self.pop()
                    elif sub_task.behavior == Behavior.Union:
                        params = list(item.bindings.values())
                        if params:
                            self.union(params[0])
                            self.interpret(sub_task)
                            self.pop()

                    elif sub_task.behavior == Behavior.OrderedChoice:
                        # print("CHOICE")
                        pass
                    else:
                        raise Exception(f"UNKNOWN SUB TASK {sub_task.behavior}")
                elif func.isa(Builtins.Install):
                    self.install(params[0])
                elif func.isa(Builtins.Quantifier):
                    self.sequence()
                    self.interpret(params[1].value)
                    if func.isa(Builtins.Not):
                        quantifier = "not"
                    elif func.isa(Builtins.Exists):
                        quantifier = "exists"
                    elif func.isa(Builtins.Every):
                        quantifier = "every"
                    else:
                        raise Exception()
                    self.quantify(quantifier, params[0].value)
                    self.pop()

                elif func.isa(Builtins.Aggregate):
                    # print("Aggregate")
                    pass
                elif func.isa(Builtins.Infix) and len(params) == 2:
                    self.filter(func.value.name, params)
                else:
                    self.compute(func.value.name, params[:-1], params[-1])
            elif item.action == ActionType.Bind and item.entity.isa(Builtins.Return):
                self.return_(item.bindings.values())
            elif item.action.is_effect():
                op = item.action.name
                props = {key.name: value for key,value in item.bindings.items()}
                self.effect(op, item.entity, [type.name for type in item.types], props)
            else:
                # print("hey!", item)
                pass



from typing import Iterable, cast
from relationalai import metamodel as mm
from relationalai.clients.test import Document

ASSERT_ACTIONS = [mm.ActionType.Bind, mm.ActionType.Persist, mm.ActionType.Unpersist]
def normalize_expected_task(expected: mm.Task|None):
    if not expected:
        return expected

    normalized = []
    ## @FIXME: Can't just fuse adjacent prior, it's with any prior
    ## @FIXME: Need to pull prev into the new one, not vice versa, since the new one might depend on things the old one didn't
    prev: mm.Action|None = None
    for ix in range(0, len(expected.items)):
        cur = expected.items[ix]
        if cur.action in ASSERT_ACTIONS:
            # Fuse adjacent asserts of the same type to the same entity
            if prev and prev.action == cur.action and prev.entity == cur.entity:
                for type in cur.types:
                    if type not in prev.types:
                        prev.types.append(type)
                for prop in cur.bindings.keys():
                    prev.bindings[prop] = cur.bindings[prop]

                continue

            prev = cur

        normalized.append(cur)

    expected.items = normalized

def normalize_binding_order(action: mm.Action, order: Iterable[mm.Property]):
    bindings = {}
    for key in order:
        bindings[key] = action.bindings.get(key, None)

    action.bindings = bindings
    return action

def map_document_ids(old: Document, new: Document):
    # @FIXME: Definitely gotta offset for pyrelstd
    for block_ix in range(min(len(old.blocks), len(new.blocks))):
        map_task_ids(old.blocks[block_ix].task, new.blocks[block_ix].task)

    return zip(old.blocks, new.blocks)

def map_task_ids(old: mm.Task|None, new: mm.Task|None):
    if not old or not new:
        return
    for item_ix in range(min(len(old.items), len(new.items))):
        map_action_ids(old.items[item_ix], new.items[item_ix])

def map_action_ids(old: mm.Action|None, new: mm.Action|None):
    if not old or not new:
        return

    if old.action != new.action:
        return

    mapped_props = zip_similar_bindings(old.bindings, new.bindings)
    map_var_ids(old.entity, new.entity)

    for old_property, new_property in mapped_props.items():
        if new_property:
            map_var_ids(old.bindings[old_property], new.bindings[new_property])
        elif not old_property.is_input:
            new.bindings[old_property] = old.bindings[old_property]

    if old.action == mm.ActionType.Get:
        normalize_binding_order(new, (cast(mm.Property, mapped_props.get(prop, prop)) for prop in old.bindings.keys()))

def map_var_ids(old: mm.Var|None, new: mm.Var|None):
    if not old or not new:
        return

    # Such hack, but if the vars aren't in all the same places
    # I think this'll always produce different results.
    new.id = old.id

def zip_similar_bindings(old: dict[mm.Property, mm.Var], new: dict[mm.Property, mm.Var]):
    return {old_prop: next((new_prop for new_prop in new.keys() if new_prop.name == old_prop.name), None) for old_prop in old.keys()}



from relationalai.debugging import Span

# For a given span name, the attribute to use as its key.
SPAN_KEYS = {
    'test': 'name',
    'benchmark': 'name',
    'run': 'idx',
    'query': 'tag',
}

def get_key(span: 'Span'):
    type = span.type
    if type in SPAN_KEYS:
        key = SPAN_KEYS[type]
        return span.attrs[key]
    return None



import time
from typing import Any, Callable, Sequence
from colorama import Style
from gentest.harness.database import get_test_hash
from gentest.util import condense_traceback, fmt_h2, fmt_indented, stringify_blocks
from hypothesis.internal.observability import TESTCASE_CALLBACKS
from relationalai.clients.test import Document, Install, Query, proxied_clients
from gentest.validate.diff import diff_to_str
from gentest.emit import DocumentEmitter
from gentest.validate.mapping import map_document_ids

from gentest.validate.errors import GenException

#===============================================================================
# Decorator
#===============================================================================

def roundtripper(fn: Any, print_given_args = False):
    key_hash = get_test_hash(fn)
    started: float = 0
    runs: int = 0

    if not print_given_args:
        fn._hypothesis_internal_print_given_args = False

    def tap_debug_logs(value: Any):
        if value["type"] == "test_case":
            nonlocal runs
            nonlocal started
            runs += 1
            if runs % 100 == 0:
                end = time.time_ns()
                elapsed = (end - started) / 1000000
                print(f"{runs} tasks tested. {elapsed:.2f}ms elapsed. Avg {elapsed / runs:.4f}ms per task")
                # pprint.pprint(value)

    def wrapper(debug = False):
        __tracebackhide__ = True
        nonlocal started
        nonlocal runs
        started = time.time_ns()
        runs = 0

        if debug:
            TESTCASE_CALLBACKS.append(tap_debug_logs)
            try:
                wrapper.inner()
            except Exception as err:
                print(fmt_err(err, None, key_hash))
            finally:
                TESTCASE_CALLBACKS.remove(tap_debug_logs)

            end = time.time_ns()
            elapsed = (end - started) / 1000000
            if runs == 0:
                runs = 1
            print(f"{runs} tasks tested. {elapsed:.2f}ms elapsed. Avg {elapsed / runs:.4f}ms per task")
        else:
            try:
                wrapper.inner()
            except Exception as err:
                setattr(err, "key_hash", key_hash)
                raise
            # try:
            #     wrapper.inner()
            # except ExceptionGroup as err:
            #     raise Exception(fmt_err(err, None, key_hash)) from None
            # except GenException as err:
            #     raise Exception(fmt_err(err, None, key_hash)) from None

    wrapper.__key_hash = key_hash
    wrapper.inner = fn
    return wrapper

def fmt_errs(errs: Sequence[Exception], key_hash: str, indent = 0):
    pad = "  " * indent
    segments = [f"{pad}{Style.BRIGHT}{len(errs)} exceptions found:{Style.NORMAL}"]
    ix = 0
    for err in errs:
        segments.append(fmt_err(err, f"{ix + 1:02}", key_hash, indent))
        ix += 1

    return "\n\n".join(segments)

def fmt_err(err: Exception, label: str|None, key_hash: str, indent = 0) -> str:
    pad = "  " * indent
    match err:
        case ExceptionGroup():
            return fmt_errs(err.exceptions, key_hash, indent)
        case GenException():
            body = err.reify().long or ""
            cut_ix = body.find("\n")
            body = body[cut_ix + 1:]

            return f"{pad}{fmt_h2(err.short, err.hash_path(key_hash), indent=indent)}\n{fmt_indented(body, indent)}"
        case _:
            return f"{pad}{fmt_h2(label, indent=indent)}\n{pad}{Style.BRIGHT}{str(err)}{Style.RESET_ALL}\n{fmt_indented(condense_traceback(err), indent + 1)}\n"


#===============================================================================
# Exec
#===============================================================================

def return_document(doc: Document):
    return doc

def drain_proxied_clients():
    proxies = proxied_clients[:]
    proxied_clients.clear()
    return proxies

def get_proxied_client():
    proxies = drain_proxied_clients()
    assert len(proxies) == 1
    return proxies[0]

def exec_and_run_callback(code: str, callback: Callable[..., Any]|None, *args: Any, ns:dict|None = None):
    if callback is None:
         callback = return_document

    try:
        exec(code, ns, None)
        return callback(get_proxied_client(), *args)
    finally:
        drain_proxied_clients()

#===============================================================================
# From code
#===============================================================================

def roundtrip_from_pyrel_code(code: str):
    # print("ORIG" + "="*76)
    # print(code)
    exec_and_run_callback(code, emit_and_compare)

def emit_and_compare(executor: Document):
    doc = DocumentEmitter()
    for block in executor.blocks:
        match block:
            case Install():
                if block.name != "pyrelstd":
                    doc.rule(block.task)
            case Query():
                doc.query(block.task)

    code2 = doc.stringify()

    print("EMIT" + "="*76)
    print(code2)

    exec_and_run_callback(code2, compare_metamodels, executor)

def compare_metamodels(new: Document, old: Document):
    map_document_ids(old, new)
    old_str = stringify_blocks(old.blocks)
    new_str = stringify_blocks(new.blocks)
    # print("OLD" + "="*77)
    # print(old_str)
    # print("NEW" + "="*77)
    # print(new_str)
    print("DIFF" + "="*76)
    print(diff_to_str(old_str, new_str))




# pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std.graphs import Graph
from relationalai.std.aggregates import sum
from relationalai.std import alias

#--------------------------------------------------
# Types
#--------------------------------------------------

model = rai.Model(
    name=globals().get("name", "test_mixed_value_types"), config=globals().get("config")
)
model._config.set("compiler.use_value_types", True)
model._config.set("compiler.use_multi_valued", True)

Person = model.Type("Person")
Criminal = model.Type("Criminal")
Transaction = model.Type("Transaction")
Dude = model.Type("Dude")

#--------------------------------------------------
# Load data
#--------------------------------------------------

with model.rule():
    joe = Person.add(name="Joe")
    jane = Person.add(name="Jane")
    bob = Person.add(name="Bob")
    dylan = Person.add(Criminal, name="Dylan")
    dave = Person.add(name="Dave")

    joe.knows.add(jane)
    jane.knows.add(bob)
    bob.knows.extend([joe, dylan])
    dylan.knows.add(joe)

    dude = Dude.add(name="Woop")
    dude.knows.add(dylan)

    Transaction.add(from_=joe, to=jane, amount=100)
    Transaction.add(from_=joe, to=jane, amount=1000)
    Transaction.add(from_=joe, to=jane, amount=10)
    Transaction.add(from_=joe, to=bob, amount=10)

#--------------------------------------------------
# Graph
#--------------------------------------------------

graph = Graph(model, with_isolated_nodes=True)
Node, Edge = graph.Node, graph.Edge

Node.extend(Person, label=Person.name)
Node.extend(Criminal, criminal=True, label=Criminal.name)
Node.extend(Dude, criminal=True, label=Dude.name)
Edge.extend(Person.knows, label="knows")
Edge.extend(Dude.knows, label="knows")

with model.rule():
    t = Transaction()
    weight = sum(t, t.amount, per=[t.from_, t.to])
    Edge.add(t.from_, t.to, label="Transfer", weight=weight)

with model.query() as select:
    n = Node()
    component = graph.compute.weakly_connected_component(n)
    Node(component)
    z = select(alias(n, "node"), alias(component, "component"))

print(z.results)


#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std import rel

model = rai.Model(name=globals().get("name", "test_numeric_outputs"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=400000000000000000000000)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

with model.query() as select:
    a = Adult()
    z = select(a.name, rel.int(128, a.age))

print(z.results)

with model.query() as select:
    z = select(rel.decimal(64, 8, rel.sqrt(2)), rel.decimal(128, 20, rel.sqrt(2)))

print(z.results)



from typing import Sequence

from .. import dsl, std

# Custom types
_String = str | dsl.Producer
_Integer = int | dsl.Producer

# NOTE: Right now, common contains all Rel stdlib relations.
# If the stdlib is split into multiple namespaces, this will have to be updated.
_str_ns = dsl.global_ns.std.common

#--------------------------------------------------
# Basic String Operations
#--------------------------------------------------

def length(string: _String) -> dsl.Expression:
    return _str_ns.string_length(string)


def lowercase(string: _String):
    return _str_ns.lowercase(string)


def uppercase(string: _String):
    return _str_ns.uppercase(string)


#--------------------------------------------------
# Split, Join, and Concatenate
#--------------------------------------------------

def split(string: _String, separator: _String) -> dsl.Expression:
    ix, part = std.Vars(2)
    _str_ns.string_split(separator, string, ix, part)
    return ix - 1, part  # Return 0-based index


def split_part(string: _String, separator: _String, index: _Integer) -> dsl.Expression:
    return _str_ns.string_split(separator, string, index + 1)  # Convert 0-based to 1-based index


def join(strings: Sequence[_String], separator: _String) -> dsl.Expression:
    model = dsl.get_graph()
    R = dsl.InlineRelation(model, list(enumerate(strings)))
    return _str_ns.string_join(separator, R)


def concat(string1: _String, string2: _String) -> dsl.Expression:
    return _str_ns.concat(string1, string2)


#--------------------------------------------------
# Substrings
#--------------------------------------------------

def contains(string: _String, substring: _String):
    return _str_ns.contains(string, substring)


def ends_with(string: dsl.Producer, suffix: _String):
    return _str_ns.ends_with(string, suffix)


def like(string: _String, pattern: _String):
    return _str_ns.like_match(pattern, string)


#--------------------------------------------------
# Find and Replace
#--------------------------------------------------

def replace(string: _String, old: _String, new: _String) -> dsl.Expression:
    return _str_ns.string_replace(string, old, new)


#--------------------------------------------------
# Regular Expressions
#--------------------------------------------------

def regex_match(string: _String, regex: _String):
    return _str_ns.regex_match(regex, string)


def regex_compile(regex: _String):
    return _str_ns.regex_compile(regex)


#--------------------------------------------------
# Exports
#--------------------------------------------------

__all__ = [
    "concat",
    "contains",
    "ends_with",
    "join",
    "length",
    "like",
    "lowercase",
    "regex_compile",
    "regex_match",
    "replace",
    "uppercase"
]



# pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std import rel
from relationalai.errors import RAIException
import pytest

model = rai.Model(name=globals().get("name", "test_out_of_context"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    p = Person()
    p.age >= 18
    woop = p.woop
    d = p.age + 10
    v = rel.sin(3)
    p.set(Adult)

with pytest.raises(RAIException):
    with model.query() as select:
        a = Adult()
        p.foo
        select(a)

with pytest.raises(RAIException):
    with model.query() as select:
        a = Adult()
        select(a, a.name, a.age, p)

with pytest.raises(RAIException):
    with model.query() as select:
        a = Adult()
        select(a, a.name, a.age, woop)

with pytest.raises(RAIException):
    with model.query() as select:
        a = Adult()
        z = d + 10
        select(a, z)

with pytest.raises(RAIException):
    with model.query() as select:
        a = Adult()
        select(a, v)


#pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_not_found"), config=globals().get("config"))
Invited = model.Type("Invited")
Invite = model.Type("Invite")
Person = model.Type("Person")

with model.rule():
    joe = Person.add(name="Joe", age=74)
    bob = Person.add(name="Bob", age=40)
    jane = Person.add(name="Jane", age=10)

    joe.friends.extend([jane, bob])
    jane.friends.extend([joe, bob])
    bob.friends.add(jane)

    Invited.add(person=joe)

with model.rule():
    p = Person()
    with model.not_found():
        Invited(person=p)

    Invite.add(person=p)

#--------------------------------------------------
# Find all friends that are older than 10 where the
# person doesn't have a friend named Jane.
# This should exlude Joe, since he's friends with Jane
#--------------------------------------------------

with model.query() as select:
    p = Person()
    with model.not_found():
        p.friends.name == "Jane"
    p.friends.age > 8
    res = select(p, p.name, p.age, p.friends.name, p.friends.age)

print(res.results)

#--------------------------------------------------
# Order matters, if we move the condition up, we'll unify
# on friends as well and keep any that don't have a name
# of Jane
#--------------------------------------------------

with model.query() as select:
    p = Person()
    p.friends.age > 8
    with model.not_found():
        p.friends.name == "Jane"
    res = select(p, p.name, p.age, p.friends.name, p.friends.age)

print(res.results)

#--------------------------------------------------
# Find all people who need an invite
#--------------------------------------------------

with model.query() as select:
    res = select(Invite().person.name)



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std import aggregates

model = rai.Model(name=globals().get("name", "test_top"), config=globals().get("config"))
Person = model.Type("Person")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=15)
    Person.add(name="Alex", age=33)
    Person.add(name="John", age=27)

with model.query() as select:
    p = Person()
    aggregates.top(3, p.age)
    select(p, p.name, p.age)



# pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std.graphs import Graph
from relationalai.std.aggregates import sum
import rich

#--------------------------------------------------
# Types
#--------------------------------------------------

model = rai.Model("MyCoolDatabase")
model._config.set("compiler.use_multi_valued", True)

Person = model.Type("Person")
Criminal = model.Type("Criminal")
Transaction = model.Type("Transaction")

#--------------------------------------------------
# Load data
#--------------------------------------------------

with model.rule():
    joe = Person.add(name="Joe")
    jane = Person.add(name="Jane")
    bob = Person.add(name="Bob")
    dylan = Person.add(Criminal, name="Dylan")
    dave = Person.add(name="Dave")

    joe.knows.add(jane)
    jane.knows.add(bob)
    bob.knows.extend([joe, dylan])
    dylan.knows.add(joe)

    Transaction.add(from_=joe, to=jane, amount=100)
    Transaction.add(from_=joe, to=jane, amount=1000)
    Transaction.add(from_=joe, to=jane, amount=10)
    Transaction.add(from_=joe, to=bob, amount=10)

#--------------------------------------------------
# Graph
#--------------------------------------------------

graph = Graph(model)
Node, Edge = graph.Node, graph.Edge

Node.extend(Person, label=Person.name)
Node.extend(Criminal, criminal=True)
Edge.extend(Person.knows, label="knows")

with model.rule():
    t = Transaction()
    weight = sum(t, t.amount, per=[t.from_, t.to])
    Edge.add(t.from_, t.to, label="Transfer", weight=weight)

with model.rule():
    n = Node()
    rank = graph.compute.pagerank(n)
    n.set(rank=rank * 5)

#--------------------------------------------------
# Go
#--------------------------------------------------

data = graph.fetch()
rich.print(data)
graph.visualize(three=True, style={
    "node": {
        "color": lambda x: "red" if x.get("criminal") else "blue",
        "size": lambda x: (x.get("rank", 1) * 2) ** 2
    },
    "edge": {
        "color": "yellow",
    }
}).display()


import sys

def uncache(exclude):
    """Remove package modules from cache except excluded ones.
    On next import they will be reloaded.

    Args:
        exclude (iter<str>): Sequence of module paths.
    """
    pkgs = []
    for mod in exclude:
        pkg = mod.split('.', 1)[0]
        pkgs.append(pkg)

    to_uncache = []
    for mod in sys.modules:
        if mod in exclude:
            continue

        if mod in pkgs:
            to_uncache.append(mod)
            continue

        for pkg in pkgs:
            if mod.startswith(pkg + '.'):
                to_uncache.append(mod)
                break

    for mod in to_uncache:
        del sys.modules[mod]



from abc import ABC
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Callable, Concatenate, ParamSpec, TypeVar, cast
from hypothesis.strategies import SearchStrategy as Strategy
from hypothesis import strategies as gen
import rich
from gentest.gen.error import BuilderConsumedError
from gentest.gen.ir import implements

from relationalai import metamodel as mm

def has_outputs(task: mm.Task) -> bool:
    for prop in task.properties:
        if not prop.is_input:
            return True
    return False

K = TypeVar("K")
V = TypeVar("V")
def get_key_for_value(obj: dict[K, V], value: V) -> K:
    for key, val in obj.items():
        if val == value:
            return key
    raise Exception(f"No key found for value {value}")

@dataclass
class GenContextBase(ABC):
    """
    GenContextBase captures the essential data used by a GenContext.
    """
    types: set[mm.Type] = field(default_factory=set)
    tasks: list[mm.Task] = field(default_factory=list)

    def __str__(self):
        console = rich.get_console()
        with console.capture() as capture:
            console.rule("Context")
            console.print("    Types")
            for type in self.types:
                console.print(f"    - {type.name} ({type.id})")
            console.print("    Tasks")
            for task in self.tasks:
                console.print(f"    - {task.name}({', '.join(f'{prop.name}: {prop.type.name}' for prop in task.properties)})")
        return capture.get()



@dataclass
class GenContext(GenContextBase):
    """
    A GenContext stores information about the world that are largely static over the duration of a document.
    """

    _sample_arbitrary_type: Strategy[mm.Type] = field(init = False)
    _sample_non_primitive_type: Strategy[mm.Type] = field(init = False)
    _sample_subtype: dict[mm.Type, Strategy[mm.Type]] = field(init = False, default_factory=dict)

    def __post_init__(self):
        self._sample_arbitrary_type = gen.sampled_from(list(self.types))
        self._sample_non_primitive_type = gen.sampled_from([t for t in self.types if not implements(t, mm.Builtins.Primitive)])

    def sample_type(self, super_type: mm.Type|None = None) -> Strategy[mm.Type]:
        if not super_type:
            return self._sample_arbitrary_type
        else:
            strat = self._sample_subtype.get(super_type)
            if not strat:
                subtypes = [t for t in self.types if implements(t, super_type)]
                if subtypes:
                    strat = gen.sampled_from(subtypes)
                    self._sample_subtype[super_type] = strat
                else:
                    raise Exception(f"No subtypes registered for supertype {super_type}")

            return strat

    def sample_non_primitive_type(self) -> Strategy[mm.Type]:
        return self._sample_non_primitive_type


    def build(self):
        return GenContextBuilder(
            types = self.types.copy(),
            tasks = self.tasks.copy())

# Create a type variable that can be any callable
P = ParamSpec("P")
R = TypeVar("R")

def copy_signature_to_method(original_function: Callable[P, R]):
    def decorator(wrapper_function: Callable) -> Callable[Concatenate[Any, P], R]:
        # Update the wrapper function to have the same signature as the original function
        return cast(Any, wraps(original_function)(wrapper_function))
    return decorator

@dataclass
class GenContextBuilder(GenContextBase):
    """
    Convenience class for constructing a GenContext. Remember that any contexts used for generation _must_ not be mutated.
    """

    finished = False

    def assert_not_finished(self):
        if self.finished:
            raise BuilderConsumedError(self)

    @copy_signature_to_method(mm.Type)
    def type(self, *args, **kwargs):
        type = mm.Type(*args, **kwargs)
        self.register_type(type)
        return type

    @copy_signature_to_method(mm.Task)
    def task(self, *args, **kwargs):
        task = mm.Task(*args, **kwargs)
        self.register_task(task)
        return task

    def register_type(self, type: mm.Type):
        self.assert_not_finished()
        self.types.add(type)

    def register_task(self, task: mm.Task):
        self.assert_not_finished()
        self.tasks.append(task)

    def finish(self) -> GenContext:
        """Efficiently drain the builder's state into a scope. Note that it _cannot_ be reused after this."""
        self.finished = True
        return GenContext(
            types = self.types,
            tasks = self.tasks)

    def snapshot(self) -> GenContext:
        """Construct a scope from the current builder state without consuming the builder."""
        return GenContext(
            types = self.types.copy(),
            tasks = self.tasks.copy())


def base_fixture(builder: GenContextBuilder):
    for k, v in vars(mm.Builtins).items():
        match v:
            case mm.Task():
                if not implements(v, mm.Builtins.Quantifier) and v != mm.Builtins.make_identity:
                    builder.register_task(v)
            case mm.Type():
                if implements(v, mm.Builtins.Primitive):
                    builder.register_type(v)

    return builder

def fixture(cb: Callable[[GenContextBuilder], Any], into: GenContextBuilder|None = None) -> GenContextBuilder:
    if not into:
        into = base_fixture(GenContextBuilder())
    cb(into)
    return into



from abc import ABC
from typing import Optional, TypedDict
from pathlib import Path
from urllib.parse import urlparse

class Import(TypedDict):
    name: str
    type: str
    hash: Optional[str]
    info: Optional[dict]
    status: Optional[str]

class AvailableModel(TypedDict):
    name: str

class ImportSource(ABC):
    name: str

class ImportSourceFile(ImportSource):
    def __init__(self, path: str, name: Optional[str] = None):
        self.raw_path = path
        self._url = urlparse(self.raw_path)
        self._path = Path(self._url.path if self.is_url() else self.raw_path)
        self.extension = self._path.suffix
        self.name = name or self._path.name

    def is_url(self):
        return bool(self._url.netloc)

class ImportSourceTable(ImportSource):
    def __init__(self, database: str, schema: str, table: str, name: Optional[str] = None):
        self.database = database
        self.schema = schema
        self.table = table
        self.name = name or self.fqn

    @property
    def fqn(self):
        return f"{self.database}.{self.schema}.{self.table}"



import logging
from typing import Dict, List

from relationalai.debugging import Span
from relationalai.util.keys import get_key

ATTR_ALLOW_LIST = {
    'transaction': {'txn_id'},
    'wait': {'txn_id'},
}

class TracingLogger(logging.Handler):
    """
    This logger prints spans to stdout as they start and finish.
    """
    
    def __init__(self):
        super().__init__()
        self.open_spans = {}

    def emit(self, record):
        if isinstance(record.msg, dict):
            msg = record.msg
            if msg['event'] == 'span_start':
                span = msg['span']
                self.open_spans[str(span.id)] = span
                path = get_path(span)
                attrs = get_attributes(span.type, span.attrs)
                print('start', format_key_path(path), attrs)
            elif msg['event'] == 'span_end':
                span = self.open_spans.pop(msg['id'])
                path = get_path(span)
                attrs = get_attributes(span.type, msg['end_attrs'])
                attrs['elapsed_s'] = (span.end_timestamp - span.start_timestamp).total_seconds()
                print('  end', format_key_path(path), attrs)


def get_path(span: Span) -> List[Span]:
    path = []
    cur = span
    while cur is not None:
        path.append(cur)
        cur = cur.parent
    path.reverse()
    return path

def get_attributes(span_type: str, attrs: Dict):
    out = {}
    allowed_attributes = ATTR_ALLOW_LIST.get(span_type, set())
    for attr in allowed_attributes:
        if attr in attrs:
            out[attr] = attrs[attr]
    return out

def format_span(span: Span):
    key = get_key(span)
    if key is None:
        return span.type
    return f"{span.type}({key})"

def format_key_path(path: List[Span]):
    return '.'.join([format_span(span) for span in path])



#pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "multi_valued"), config=globals().get("config"))
model._config.set("compiler.use_multi_valued", True)

Person = model.Type("Person")
Adult = model.Type("Adult")

Person.foos.has_many()

with model.rule():
    p = Person.add(name="Joe", age=74)
    p.foos.extend([8,9,0])
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    foos = p.foos
    p.age >= 18
    foos.extend([1,2,3])
    p.age.in_([10,20,40,74])
    p.set(Adult)

# This rule makes sure that we aren't incorrectly removing the get
# for a collection if we do an append on the same property instance
with model.rule():
    p = Person()
    foos = p.foos
    foos > 8
    foos.add(11)

with model.query() as select:
    p = Person()
    p.foos.in_([8,9,0])
    z = select(p.name)

print(z.results)

with model.query() as select:
    p = Person()
    z = select(p.name, p.foos)

print(z.results)

with model.query() as select:
    p = Person(name="Joe")
    a, b, c = p.foos.choose(3)
    z = select(p.name, a, b, c)

print(z.results)

with model.query() as select:
    p = Person()
    p2 = Person()
    p < p2
    p.foos == p2.foos
    z = select(p.name, p2.name)

print(z.results)

with model.query() as select:
    a = Adult()
    z = select(a, a.name, a.age)

print(z.results)


import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as these lists.
NO_ARGS = [
    "avg_clustering_coefficient",
]

ONE_ARG = [
    "local_clustering_coefficient",
]

model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    obj3 = Object.add(id=3)
    obj4 = Object.add(id=4)
    Relationship.add(from_=obj1, to=obj2)
    Relationship.add(from_=obj1, to=obj3)
    Relationship.add(from_=obj1, to=obj4)
    Relationship.add(from_=obj2, to=obj3)

graph = std.graphs.Graph(model, undirected=True)
graph.Node.extend(Object)

with model.rule():
    r = Relationship()
    graph.Edge.add(r.from_, r.to)

for algo_name in NO_ARGS:
    with model.query(tag=algo_name) as select:
        algo = getattr(graph.compute, algo_name)()
        select(algo)

for algo_name in ONE_ARG:
    with model.query(tag=algo_name) as select:
        o = Object()
        algo = getattr(graph.compute, algo_name)(o)
        select(o, algo)


import time
import relationalai as rai
from relationalai.std import rel
from relationalai.clients.snowflake import Snowflake
from relationalai.std import aggregates

model = rai.Model("IMDB_EXAMPLE")

start = time.time()

sf = Snowflake(model)

Title = sf.imdb_example.imdb.titles

with model.query() as select:
    t = Title()
    rel.regex_match(t.genres, "/Documentary/") # Only show the documentaries
    t.startyear > 1915 # Only show the movies after 1915
    aggregates.rank_desc(t.primarytitle) <= 25 # Only show the top 25 movies
    response = select(t.primarytitle, t.genres, t.startyear) # Select the columns to show

print(response.results)

end = time.time()

print("\nDone in {:.2f}s".format(end - start))



sf_profile = {}
sf_profile['platform'] = 'snowflake'
sf_profile['user'] = 'usr'
sf_profile['password'] = 'pwd'
sf_profile['warehouse'] = 'WH'
sf_profile['role'] = 'CONSUMER'
sf_profile['rai_app_name'] = 'APP'
sf_profile['account'] = 'foo-account'

az_profile = {}
az_profile['platform'] = 'azure'
az_profile['client_id'] = 'cl_id'
az_profile['client_secret'] = 'cl_s'
az_profile['client_credentials_url'] = '123456'
az_profile['host'] = 'azure-latest.relationalai.com'
az_profile['port'] = 443
az_profile['region'] = 'us-east'
az_profile['scheme'] = 'https'

mocked_import_stream_response = [{
    'id': 'foo_id',
    'created': '2024-06-03',
    'created_by': 'JOE',
    'status': 'CREATED',
    'reference_name': 'DATA_STREAM_TABLE',
    'reference_alias': '2644cadb-5e8c-469f-95aa-f91c5d50874a',
    'fq_object_name': 'FOO',
    'rai_database': 'FOO',
    'rai_relation': 'imdb_example_imdb_titles',
    'data_sync_status': 'SYNCED',
    'pending_batches_count': 0,
    'next_batch_status': None,
    'next_batch_unloaded_timestamp': None,
    'next_batch_details': None,
    'last_batch_details': '{\n  "rows": 10000,\n  "size": 466459,\n  "writeChangesDuration": 2820,\n  "writeChangesEnd": "2024-06-03T17:51:22.888Z",\n  "writeChangesStart": "2024-06-03T17:51:20.068Z"\n}',
    'last_batch_unloaded_timestamp': '2024-06-03',
    'cdc_status': 'STARTED'
}]

mocked_list_imports_response = [{
    'id': 'foo_id',
    'model': 'FOO',
    'name': 'FOO',
    'created': '2024-06-03',
    'creator': 'JOE',
    'status': 'LOADED',
    'errors': None,
    'batches': '1'
},
{
    'id': 'bar_id',
    'model': 'BAR',
    'name': 'BAR',
    'created': '2024-01-01',
    'creator': 'JAKE',
    'status': 'LOADED',
    'errors': None,
    'batches': '1'
},
{
    'id': 'baz_id',
    'model': 'BAZ',
    'name': 'BAZ',
    'created': '2024-03-03',
    'creator': 'BLAKE',
    'status': 'LOADED',
    'errors': 'Bad data',
    'batches': '1'
}]

mocked_imports_get_details = [
    "Field                           Value",
    "id                              foo_id",
    "created                         2024-06-03",
    "created_by                      JOE",
    "status                          CREATED",
    "reference_name                  DATA_STREAM_TABLE",
    "reference_alias                 2644cadb-5e8c-469f-95aa-f91c5d50874a",
    "fq_object_name                  FOO",
    "rai_database                    FOO",
    "rai_relation                    imdb_example_imdb_titles",
    "data_sync_status                SYNCED",
    "pending_batches_count           0",
    "next_batch_status               N/A",
    "next_batch_unloaded_timestamp   N/A",
    "next_batch_details              N/A",
    """last_batch_details              {                                             
                                    "rows": 10000,                              
                                    "size": 466459,                             
                                    "writeChangesDuration": 2820,               
                                    "writeChangesEnd":                          
                                  "2024-06-03T17:51:22.888Z",                   
                                    "writeChangesStart":                        
                                  "2024-06-03T17:51:20.068Z"                    
                                  }""",
    "last_batch_unloaded_timestamp   2024-06-03",
    "cdc_status                      STARTED"
]

mocked_imports_status = {'engine': 'FOO_ENG', 'status': 'started', 'info': '{\n  "createdOn": "2024-06-03 10:50:04.735 -0700",\n  "lastSuspendedOn": null,\n  "lastSuspendedReason": null,\n  "state": "started"\n}'}

mocked_engines_list = [
    {'name': 'test', 'size': 'XS', 'state': 'READY'},
    {'name': 'foo', 'size': 'XS', 'state': 'READY'},
    {'name': 'bar', 'size': 'S', 'state': 'READY'},
    {'name': 'baz', 'size': 'M', 'state': 'PENDING'},
    {'name': 'goo', 'size': 'M', 'state': 'SUSPENDED'},
]



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std import rel, Vars

model = rai.Model("MyCoolDatabase")
model.load_raw("rel/bar.rel")
model.load_raw("rel/foo.rel")

with model.query() as select:
    a,b = Vars(2)
    rel.foo(a, b)
    rel.bar(a + 5)
    z = select(a, b)

print(z.results)









from relationalai import metamodel as mm
from gentest.gen.context import GenContextBuilder

def person_place_thing(builder: GenContextBuilder):
    std = mm.Builtins
    place = builder.type("Place", [
        mm.Property("address", std.String),
        mm.Property("zip", std.Int),
        mm.Property("lat", std.Decimal),
        mm.Property("lng", std.Decimal),
    ])
    thing = builder.type("Thing", [
        mm.Property("name", std.String),
        mm.Property("at", place),
    ])
    person = builder.type("Person", [
        mm.Property("name", std.String),
        mm.Property("age", std.Number),
        mm.Property("height", std.Number),
        mm.Property("home", place),
        mm.Property("owns", thing),
    ])
    person.properties.append(mm.Property("friends", person))



import requests
import json

URL = "http://eventregistry.org/api/v1/article/getArticles"
KEYWORDS = [
    ["Municipal Bonds", "Bond", "Municipal"],
    ["Finance", "Business"],
    ["Stock", "Equity", "Share"],
    ["Technology", "Science"],
    ["Sports"],
    ["Politics"],
]

node_sizes = {
    "article": 45,
    "source": 40,
    "language": 35,
    "author": 30,
    "people": 25,
    "location": 20,
    "organization": 15,
}

node_colors = {
    "article": "blue",
    "source": "green",
    "language": "yellow",
    "author": "pink",
    "people": "cyan",
    "location": "brown",
    "organization": "purple",
}

edge_colors = {
    "publishedBy": "grey",
    "writtenIn": "purple",
    "writtenBy": "orange",
    "people": "blue",
    "location": "green",
    "organization": "red",
}


# --------------------------------------------------
# Functions to get articles
# --------------------------------------------------
def get_articles_request(keywords, num_articles):
    print("[+] Getting articles for keywords:", keywords)
    articles = []

    api_key = "YOUR_API_KEY"

    payload = {
        "action": "getArticles",
        "keyword": keywords,
        "articlesPage": 1,
        "articlesSortBy": "date",
        "articlesSortByAsc": False,
        "articlesArticleBodyLen": -1,
        "resultType": "articles",
        "isDuplicateFilter": "skipDuplicates",
        "dataType": "news",
        "apiKey": api_key,
        "forceMaxDataTimeWindow": 31,
        "articlesCount": num_articles,
    }

    headers = {"Content-Type": "application/json"}

    response = requests.request("POST", URL, headers=headers, data=json.dumps(payload))
    try:
        articles = json.loads(response.text.encode("utf8"))
    except Exception as e:
        print("Exception occured while getting articles", e)
        return list()

    return articles["articles"]["results"]


def get_articles(num_articles=100, use_local=True, keywords=KEYWORDS[3]):
    date = "04_04_2024"
    file_name = "daily_articles/" + date + ".json"

    if not use_local:
        print(f"\tGetting articles from API for keywords: {keywords}")
        articles = list()
        for sub_keywords in keywords:
            articles += get_articles_request(sub_keywords, num_articles)
        return articles
    else:
        print(f"\tFound articles at: {file_name}")
        with open(file_name, "r") as handle:
            data_from_json = json.load(handle)
            articles = data_from_json["articles"]["results"][0:num_articles]
            return articles


##--------------------------------------------------
# Helper Functions
# --------------------------------------------------
def get_node_size_by_type(n):
    return node_sizes.get(n.get("type"), 10)


def get_node_color_by_type(n):
    return node_colors.get(n.get("type"), "red")


def get_edge_color_by_type(e):
    return edge_colors.get(e.get("type"), "black")


def get_author(article):
    authors = article.get("authors")
    if authors and isinstance(authors, list) and len(authors) > 0:
        first_author_name = authors[0].get("name", "Unknown")
    else:
        first_author_name = "Unknown Author"
    return first_author_name


def trim_node_label(label, word_limit=3):
    words = label.split()

    trimmed_label = " ".join(words[:word_limit]) + (
        "..." if len(words) > word_limit else ""
    )
    return trimmed_label



import json
import sys
import time
import threading
import subprocess
from typing import cast
import click
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
import os
import rich
import multiprocessing
from relationalai import metagen


from .cli_controls import divider, Spinner

#--------------------------------------------------
# Root
#--------------------------------------------------

@click.group()
def cli():
    pass

#--------------------------------------------------
# Watch
#--------------------------------------------------

def clear():
    os.system('cls' if os.name == 'nt' else 'clear')

class ChangeHandler(PatternMatchingEventHandler):
    patterns = ["*.py", "*.rel", "raiconfig.toml"]

    def __init__(self):
        super().__init__()
        self.script = None
        self.process = None
        self.event_lock = threading.Lock()
        self.has_queued_events = False

    def check_event(self, event):
        if not event.src_path.endswith(".py"):
            return

        if "examples/" in event.src_path or "tests/end2end/test_cases/" in event.src_path:
            self.script = event.src_path

    def on_any_event(self, event):
        self.check_event(event)
        with self.event_lock:
            if self.process and self.process.poll() is None:
                # Mark that there are queued events
                self.has_queued_events = True
            else:
                self.start_process()

    def start_process(self):
        if self.script is None:
            return

        clear()
        rich.print(f"[yellow bold]{os.path.basename(self.script)}")
        rich.print("[yellow]------------------------------------------------------")
        rich.print("")

        # Start or restart the script
        self.process = subprocess.Popen(['python', self.script], shell=False)
        # Use a thread to wait for the process to finish without blocking
        wait_thread = threading.Thread(target=self.wait_and_restart_if_needed)
        wait_thread.start()

    def wait_and_restart_if_needed(self):
        self.process.wait()

        with self.event_lock:
            if self.has_queued_events:
                # Reset the flag and restart the process for batched events
                self.has_queued_events = False
                # Delay added to allow for potentially more events to accumulate
                # Adjust or remove the delay as needed
                time.sleep(0.5)
                self.start_process()

@cli.command()
@click.argument('directory', type=click.Path(exists=True))
def watch(directory):
    """Watch a DIRECTORY and re-run a SCRIPT on file changes."""
    event_handler = ChangeHandler()
    observer = Observer()
    observer.schedule(event_handler, path=directory, recursive=True)  # Now recursive
    observer.start()

    clear()
    rich.print(f"[yellow]Watching for changes in '{directory}'.")
    rich.print("[yellow]Save a script in examples/ to run it.")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()

#--------------------------------------------------
# Code stats
#--------------------------------------------------

def cloc(paths):
    script_dir = os.path.dirname(os.path.realpath(__file__))
    src_dir = os.path.join(script_dir, '..')
    core_paths = [os.path.join(src_dir, f) for f in paths]
    process = subprocess.Popen(['cloc', '--json', *core_paths], stdout=subprocess.PIPE, text=True)
    output, _ = process.communicate()

    try:
        # Parse the JSON output
        data = json.loads(output)
        # The total number of source lines is under 'SUM'/'code'
        total_lines = data['SUM']['code'] if 'SUM' in data else 0
        return cast(int, total_lines)
    except Exception as e:
        print(f"Error while parsing cloc output: {e}")
        return 0

@cli.command()
def stats():
    dsl = cloc(["dsl.py"])
    compiler = cloc(["compiler.py"])
    metamodel = cloc(["metamodel.py"])
    emitter = cloc(["rel_emitter.py"])
    rel = cloc(["rel.py"])
    metagen = cloc(["metagen.py"])
    gentest = cloc(["../gentest", "metagen.py"])
    tools = cloc(["tools/"])
    clients = cloc(["clients/"])
    std = cloc(["std/"])
    non_test_total = cloc(["."]) - metagen
    total = non_test_total + gentest
    core = dsl + compiler + metamodel + emitter + rel

    max_width = len(f"{total:,}")

    # Print statements with numbers right-aligned
    divider()
    rich.print(f"[yellow]RelationalAI  {non_test_total:>{max_width},} loc")
    rich.print(f"[yellow]  Core        {core:>{max_width},} loc")
    rich.print(f"[yellow]    dsl       {dsl:>{max_width},} loc")
    rich.print(f"[yellow]    rel       {rel:>{max_width},} loc")
    rich.print(f"[yellow]    emitter   {emitter:>{max_width},} loc")
    rich.print(f"[yellow]    metamodel {metamodel:>{max_width},} loc")
    rich.print(f"[yellow]    compiler  {compiler:>{max_width},} loc")
    rich.print(f"[yellow]  Clients     {clients:>{max_width},} loc")
    rich.print(f"[yellow]  Std         {std:>{max_width},} loc")
    rich.print(f"[yellow]  Tools       {tools:>{max_width},} loc")
    rich.print("")
    rich.print(f"[cyan]Gentest       {gentest:>{max_width},} loc")
    rich.print("")
    rich.print(f"[magenta]All           {total:>{max_width},} loc")
    divider()


#--------------------------------------------------
# Metagen
#--------------------------------------------------

@cli.command("gen")
@click.option('--total', default=50000, help='Total number of models to generate.')
@click.option('--threads', default=multiprocessing.cpu_count(), help='Threads to use, default is CPU count.')
@click.option('--internal', default=False, is_flag=True)
def gen(total, threads, internal):
    if not internal:
        divider()
    with Spinner(f"Testing {total:,.0f} models on {threads:,.0f} threads", f"Tested {total:,.0f} models"):
        if threads > 1:
            (elapsed, results) = metagen.batches(total, threads)
        else:
            gen = metagen.batch(total)
            results = [gen]
            elapsed = gen.elapsed

    rich.print("")

    for result in results:
        rich.print(result)

    failed = False
    for result in results:
        if len(result.failures) > 0:
            failed = True
            rich.print()
            result.print_failures(1)
            rich.print("")
            break

    rich.print("")
    rich.print(f"[yellow bold]Total time: {elapsed:,.3f}s")
    rich.print("")
    if not internal:
        divider()
        sys.exit(failed)

#--------------------------------------------------
# Metagen Watch
#--------------------------------------------------

class MetagenWatcher(PatternMatchingEventHandler):
    patterns = ["*.py"]

    def __init__(self, total, threads):
        super().__init__()
        self.process = None
        self.event_lock = threading.Lock()
        self.has_queued_events = False
        self.total = total
        self.threads = threads
        self.start_process()

    def on_any_event(self, event):
        with self.event_lock:
            if self.process and self.process.poll() is None:
                # Mark that there are queued events
                self.has_queued_events = True
            else:
                self.start_process()

    def start_process(self):
        clear()
        rich.print("[yellow bold]Metagen")
        rich.print("[yellow]------------------------------------------------------")
        rich.print("")

        # Start or restart the script
        self.process = subprocess.Popen(['rai-dev', 'gen', '--total', str(self.total), '--threads', str(self.threads), "--internal"], shell=False)
        # Use a thread to wait for the process to finish without blocking
        wait_thread = threading.Thread(target=self.wait_and_restart_if_needed)
        wait_thread.start()

    def wait_and_restart_if_needed(self):
        self.process.wait()

        with self.event_lock:
            if self.has_queued_events:
                # Reset the flag and restart the process for batched events
                self.has_queued_events = False
                # Delay added to allow for potentially more events to accumulate
                # Adjust or remove the delay as needed
                time.sleep(0.5)
                self.start_process()


@cli.command("gen:watch")
@click.argument('directory', type=click.Path(exists=True))
@click.option('--total', default=20000, help='Total number of models to generate')
@click.option('--threads', default=multiprocessing.cpu_count(), help='Threads to use')
def gen_watch(directory, total, threads):
    clear()
    event_handler = MetagenWatcher(total, threads)
    observer = Observer()
    observer.schedule(event_handler, path=directory, recursive=True)  # Now recursive
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()


import relationalai as rai
from relationalai.std import alias

model = rai.Model('EVPenetrationExample')
State = model.Type('State')

# Add some data to the collection
with model.rule():
    State.add(name='California', population=39_512_223, ev_registration=425_300)
    State.add(name="Delaware", population=973_764, ev_registration=1950)
    State.add(name="Missouri", population=6_137_428, ev_registration=6740)

# Query the collection
with model.query() as select:
    state = State()
    ev_penetration = 1000 * state.ev_registration / state.population
    response = select(
        state.name,
        state.population,
        state.ev_registration,
        alias(ev_penetration, "ev_penetration"),
    )

# Display the results sorted in descending order by the `ev_penetration` column
print(response.results.sort_values(by="ev_penetration", ascending=False))



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.clients.snowflake import Snowflake
from relationalai.std import aggregates, rel
from relationalai.std.graphs import Graph

from relationalai import debugging

# This benchmark expects the model to already be present, with the data loaded into it.
model = rai.Model("PyRelTastyBytesBenchmark", config=globals().get("config"))
model._config.set("compiler.use_multi_valued", True)
sf = Snowflake(model)
Record = sf.TASTYBYTES.HARMONIZED.LOYALTY_ORDERS_REGION_CALIFORNIA

Customer = model.Type("Customer")
Truck = model.Type("Trucks")
Transaction = model.Type("Transaction")
RelevantConnection = model.Type("RelevantConnection")

with model.query(tag="count_all") as select:
    record = Record()
    num_records = aggregates.count(record)
    select(num_records)

# Define Customer Type
with model.rule(dynamic=True):
    r = Record()
    Customer.add(customer_id=r.customer_id)

# Check total number of customers
with model.query(tag="count_customers") as select:
    record = Customer()
    num_records = aggregates.count(record)
    select(num_records)

# Define Truck Type
with model.rule():
    r = Record()
    Truck.add(truck_id=r.truck_id)

# Check total number of trucks
with model.query(tag="count_trucks") as select:
    record = Truck()
    num_records = aggregates.count(record)
    select(num_records)

with model.rule():
    r = Record()
    Transaction.add(
        customer_id=r.customer_id,
        order_id=r.order_id,
        truck_id=r.truck_id,
        order_ts=r.order_ts,
        order_ts_seconds=r.order_ts_seconds,
        location_id=r.location_id,
    )

with model.rule():
    t1 = Transaction()
    t2 = Transaction()

    t1.truck_id == t2.truck_id
    t1.customer_id != t2.customer_id
    rel.abs(t1.order_ts_seconds - t2.order_ts_seconds) <= 1200

    t1.connected.add(t2)

with model.query(tag="count_connected_customers") as select:
    t = Transaction()
    num_records = aggregates.count(t.customer_id, t.order_ts, t.connected, t.connected.customer_id)
    select(num_records)

with model.rule():
    t = Transaction()
    total_connections = aggregates.count(
        t, per=[t.customer_id, t.connected.customer_id]
    )
    total_connections > 4
    RelevantConnection.add(
        customer_1=Customer(customer_id=t.customer_id),
        customer_2=Customer(customer_id=t.connected.customer_id),
        total_connections=total_connections,
    )

# Get the total occurrences where pairs of customers coexisted together more than once
with model.query(tag="count_meaningful_connected_customers") as select:
    record = RelevantConnection()
    num_records = aggregates.count(record)
    select(num_records)

community_graph = Graph(model, undirected=True)

# Add edges to the graph between customers / Nodes will be added automatically
with model.rule():
    connection = RelevantConnection()
    community_graph.Edge.add(
        connection.customer_1,
        connection.customer_2,
        weight=connection.total_connections,
    )

with model.rule():
    customer = Customer()
    community_id = community_graph.compute.louvain(customer)
    customer.set(community_id=community_id)

    community_graph.Node.add(
        customer,
        community_id=community_id,
        customer_id=customer.customer_id
    )

with debugging.span("fetch", graph="louvain"):
    community_graph.fetch()


# pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_persist2"), config=globals().get("config"))
PersistFoo = model.Type("PersistFoo")

with model.query():
    PersistFoo.persist(p_name="Foo2")
    PersistFoo.persist(p_name="Foo3")

with model.query() as select:
    p = PersistFoo()
    z = select(p, p.p_name)

print("Persists!\n")
print(z.results)

with model.query():
    p = PersistFoo()
    p.unpersist(p_name=p.p_name)

with model.query() as select:
    p = PersistFoo()
    z = select(p, p.p_name)

print("\nUnpersists!\n")
print(z.results)



import os
from typing import Callable
from hypothesis.database import _hash, DirectoryBasedExampleDatabase
from hypothesis.internal.reflection import function_digest
from gentest.harness.vendor_types import HypothesisTestFn
from gentest.util import PROJECT_DIR


HYPO_DB_PATH = os.path.join(os.path.dirname(PROJECT_DIR), ".hypothesis", "examples")

HYPO_DB = DirectoryBasedExampleDatabase(HYPO_DB_PATH)
KEY_LEN = 8
VAL_LEN = 10

def get_test_hash(test_fn: Callable|HypothesisTestFn):
    if hasattr(test_fn, "is_hypothesis_test"):
        return get_test_inner_hash(test_fn.hypothesis.inner_test)
    raise Exception("Must pass a valid hypothesis test fn to get its key. Did you mean to pass the inner fn to `get_test_inner_key()`?")


def get_test_inner_hash(test_inner_fn: Callable):
    return _hash(function_digest(test_inner_fn))

def get_value_hash(value: bytes):
    return _hash(value)

def get_short_path(key_hash: str, value_hash: str):
    return f"{key_hash[:KEY_LEN]}/{value_hash[:VAL_LEN]}"

def expand_short_key(short_key: str, db_path: str):
    if not os.path.isdir(db_path):
        raise Exception("No hypothesis database detected.")

    keys = os.listdir(db_path)
    key = None
    for cur_key in keys:
        if cur_key.startswith(short_key):
            if key is not None:
                raise Exception("Hash collision! Specify a longer segment of the key hash.")
            key = cur_key

    if key:
        return key
    else:
        raise Exception(f"Invalid test key hash: '{short_key}'")

def expand_short_value(key: str, short_value: str, db_path: str):
    key_path = os.path.join(db_path, key)
    if not os.path.isdir(key_path):
        raise Exception(f"No directory found for key '{key}'.")

    values = os.listdir(key_path)
    value = None
    for cur_value in values:
        if cur_value.startswith(short_value):
            if value is not None:
                raise Exception("Hash collision! Specify a longer segment of the value hash.")
            value = cur_value

    if value:
        return value
    else:
        raise Exception(f"Invalid repro input value hash: '{short_value}'")

# Function to retrieve a particular failing input by hash ID
def expand_short_path(short_path: str, db_path: str = HYPO_DB_PATH, relative = False):
    (short_key, short_value) = short_path.split("/")
    if not short_key or not short_value:
        raise Exception("Malformed repro short path. It should be in the form <key_hash>/<value_hash>")

    key = expand_short_key(short_key, db_path)
    value = expand_short_value(key, short_value, db_path)

    if relative:
        return os.path.join(key, value)
    else:
        return os.path.join(db_path, key, value)

def get_conjecture_data(file_path: str):
    with open(file_path, "rb") as file:
        conjecture_data = file.read()
        return conjecture_data



#pyright: reportUnusedExpression=false

import relationalai as rai
import streamlit as st
import pandas as pd

model = rai.Model("MyCoolDatabase")
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=54)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age > 18
    p.set(Adult)

min_age = st.number_input("Age", min_value=0, max_value=100, value=18, step=1, key="age")

with st.spinner("Running query..."):
    with model.query() as select:
        a = Adult()
        a.age > min_age
        z = select(a, a.name, a.age)

st.header("Adults above a certain age", divider=True)
st.table(pd.DataFrame(z.results, columns=["id", "name", "age"]))






from pathlib import Path
from tests.util import all_files_in_dir, do_snapshot_test
from relationalai.clients import config as cfg

@all_files_in_dir(Path(__file__).parent / "benchmarks")
def test_snapshots(file_path: Path, snapshot, engine_config: cfg.Config, engine_size):
    do_snapshot_test(
        engine_config, engine_size, snapshot, file_path, span_name="benchmark",
        create_db=False, runs=2,
    )



import pytest

import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as this list.
ALGOS = ["num_nodes", "num_edges"]

model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    Relationship.add(from_=obj1, to=obj2)

graph = std.graphs.Graph(model)

# Make sure 'undirected' is set to False by default.
assert not graph.undirected

# Make sure 'weighted' is set to False by default.
assert not graph.weighted

# Make sure trying to set 'undirected' to True raises an error.
with pytest.raises(AttributeError):
    graph.undirected = True

# Make sure trying to set 'weighted' to True raises an error.
with pytest.raises(AttributeError):
    graph.weighted = True

# Add edges to the graph.
with model.rule():
    r = Relationship()
    graph.Edge.add(r.from_, r.to)

# Test each algorithm.
for algo_name in ALGOS:
    with model.query(tag=algo_name) as select:
        o = Object()
        # NOTE: This only works for unary relations.
        algo = getattr(graph.compute, algo_name)()
        select(algo)



import ast
import contextlib
from dataclasses import dataclass
import datetime
import inspect
import io
import json
import sys
from textwrap import dedent
import os
from types import TracebackType
from typing import Dict
import logging
import uuid
import warnings

import numpy as np

from pandas import DataFrame

from relationalai.errors import InvalidIfWarning, InvalidLoopWarning, InvalidTryWarning, RAIException, RAIWarning
from .metamodel import Action, Task

DEBUG = True
handled_error = None

#--------------------------------------------------
# Global Warning Handling
#--------------------------------------------------

# Save the original warnings.showwarning
original_showwarning = warnings.showwarning

def global_warning_handler(message, category, filename, lineno, file=None, line=None):
    if isinstance(message, RAIWarning):
        message.pprint()
    else:
        # Use the original showwarning for non-RAIWarning warnings
        original_showwarning(message, category, filename, lineno, file, line)

def install_warninghook():
    # Override the default showwarning with the custom one
    warnings.showwarning = global_warning_handler

#--------------------------------------------------
# Global Error Handling
#--------------------------------------------------

def global_exception_handler(exc_type, exc_value, exc_traceback, quiet=False):
    global handled_error

    if isinstance(exc_value, RAIException):
        exc_value.pprint()

    # Ensure exc_value is actually an exception instance
    if isinstance(exc_value, BaseException):
        error(exc_value)

    handled_error = None

    if original_excepthook is not None:
        original_excepthook(exc_type, exc_value, exc_traceback)

original_excepthook = None

def ipy_exception_handler(shell, exc_type, exc_value, exc_traceback: TracebackType | None, tb_offset=1):
    with contextlib.redirect_stdout(io.StringIO()) as buffer:
        global_exception_handler(exc_type, exc_value, exc_traceback, True)

    if isinstance(exc_value, RAIException):
        pprinted = buffer.getvalue().strip()
        # remove the trailing divider
        pprinted = pprinted[:pprinted.rfind("\n")]
        # We strip the most _recent_ frame (rather than the top frame) because it's always part of our error handling
        exc_traceback = drop_last_frame(exc_traceback)
        exc_value = Exception(exc_value.name).with_traceback(exc_traceback)

        print(pprinted, file=sys.stderr)

    # @NOTE: We could completely takeover display for RAIExceptions to
    #        have finer control over the display, but I went with the
    #        less invasive approach for now.
    shell.showtraceback((exc_type, exc_value, exc_traceback), tb_offset=tb_offset)

def setup_exception_handler():
    try:
        from IPython import get_ipython  # type: ignore

        ipython = get_ipython()
        if ipython:
            ipython.set_custom_exc((BaseException,), ipy_exception_handler)
            return
    except Exception:
        pass

    global original_excepthook
    original_excepthook = sys.excepthook
    sys.excepthook = global_exception_handler


def drop_last_frame(tb: TracebackType | None):
    if tb is None or tb.tb_next is None:
        return None

    tail = drop_last_frame(tb.tb_next)
    return TracebackType(tail, tb.tb_frame, tb.tb_lasti, tb.tb_lineno)

#--------------------------------------------------
# Log Formatters
#--------------------------------------------------

def encode_log_message(obj):
    if isinstance(obj, DataFrame):
        # Replace NaN with None to avoid JSON serialization issues
        df = obj.replace({np.nan: None})
        return df.head(20).to_dict(orient="records")
    if isinstance(obj, datetime.datetime):
        return obj.isoformat() + "Z"
    if hasattr(obj, "to_dict"):
        return obj.to_dict()
    if hasattr(obj, "to_json"):
        return obj.to_json()
    else:
        return str(obj)

class JsonFormatter(logging.Formatter):
    def format(self, record):
        return json.dumps(record.msg, default=encode_log_message)

#--------------------------------------------------
# Logging
#--------------------------------------------------

logger = logging.getLogger("pyrellogger")
logger.setLevel(logging.DEBUG)
logger.propagate = False

#--------------------------------------------------
# File Logger
#--------------------------------------------------

class FlushingFileHandler(logging.FileHandler):
    def emit(self, record):
        super().emit(record)
        self.flush()

with open('debug.jsonl', 'w'):
    pass

# keep the old file-based debugger around and working until it's fully replaced.
if DEBUG:
    file_handler = FlushingFileHandler('debug.jsonl', mode='a')
    file_handler.setFormatter(JsonFormatter())
    logger.addHandler(file_handler)

#--------------------------------------------------
# Test Logger
#--------------------------------------------------

# class TestHandler(logging.Handler):
#     def emit(self, record):
#         d = record.msg
#         if not isinstance(d, dict):
#             return

#         print(d["event"], d.get("span", None), d.get("elapsed", None))

# if DEBUG:
#     logger.addHandler(TestHandler())

#--------------------------------------------------
# Debug Spans
#--------------------------------------------------

# The deepest span in the tree
TIP_SPAN: 'Span | None' = None

def get_current_span_id() -> str | None:
    if TIP_SPAN is None:
        return None
    return str(TIP_SPAN.id)

def span_start(type: str, **kwargs) -> 'Span':
    global TIP_SPAN
    span = Span(type, TIP_SPAN, kwargs)
    TIP_SPAN = span

    if DEBUG:
        logger.debug({"event": "span_start", "span": span})

    return span

def span_end(span):
    if not DEBUG:
        return

    global TIP_SPAN
    TIP_SPAN = span.parent
    span.mark_finished()

    logger.debug({
        "event": "span_end",
        "id": str(span.id),
        "end_timestamp": span.end_timestamp.isoformat() + "Z",
        "end_attrs": span.end_attrs,
    })

def span_flush():
    while TIP_SPAN:
        span_end(TIP_SPAN)

class Span:
    def __init__(self, type: str, parent, attrs: Dict):
        self.id = uuid.uuid4()
        self.parent = parent
        self.type = type
        self.attrs = attrs
        # additional attributes added during the lifetime of the span
        self.end_attrs = {}
        # Would use `datetime.datetime.now(datetime.UTC)`, but it only works in 3.11.
        self.start_timestamp = datetime.datetime.utcnow()
        self.end_timestamp = None

    def mark_finished(self):
        self.end_timestamp = datetime.datetime.utcnow()

    def __setitem__(self, key, value):
        self.end_attrs[key] = value

    def to_json(self):
        return {
            "type": self.type,
            "id": str(self.id),
            "parent_id": str(self.parent.id) if self.parent else None,
            "start_timestamp": self.start_timestamp.isoformat() + "Z",
            "end_timestamp": None if self.end_timestamp is None else (self.end_timestamp.isoformat() + "Z"),
            "attrs": self.attrs,
        }

@contextlib.contextmanager
def span(type: str, **kwargs) -> Span:
    cur = span_start(type, **kwargs)
    try:
        yield cur
    except Exception as err:
        error(err)
        raise
    finally:
        span_end(cur)

def set_span_attr(attr, value):
    global TIP_SPAN
    TIP_SPAN.attrs[attr] = value

#--------------------------------------------------
# Debug Events
#--------------------------------------------------

EMPTY = {}

def event(event:str, parent:Span|None = None, **kwargs):
    if not DEBUG:
        return

    if not parent:
        parent = TIP_SPAN

    d = {
        "event":event,
        "timestamp": datetime.datetime.utcnow(),
        "parent_id": parent.id if parent else None,
        **kwargs
    }
    logger.debug(d)

def time(type:str, elapsed:float, results:DataFrame = DataFrame(), **kwargs):
    if DEBUG:
        event("time", type=type, elapsed=elapsed, results={
            "values": results,
            "count": len(results)
        }, **kwargs)

def error(err:Exception):
    global handled_error
    if err != handled_error:
        # Prepare kwargs from exception attributes if it's a custom exception
        kwargs = {}

        if hasattr(err, '__dict__'):
            kwargs = {key: getattr(err, key) for key in err.__dict__ if not key.startswith('_')}

        event("error", err=err, span_id=get_current_span_id(), **kwargs)  # Emit the error event only if it's a new or different error
        for handler in logger.handlers:
            handler.flush()
        handled_error = err

def handle_compilation(compilation):
    if not DEBUG:
        return

    (file, line, block) = compilation.get_source()
    source = {"file": file, "line": line, "block": block}
    passes = [{"name": p[0], "task": p[1], "elapsed": p[2]} for p in compilation.passes]
    emitted = compilation.emitted
    if isinstance(emitted, list):
        emitted = "\n\n".join(emitted)

    event("compilation", source=source, task=compilation.task, passes=passes, emitted=emitted, emit_time=compilation.emit_time)


def warn(warning: Warning):
    if not DEBUG:
        return

    kwargs = {}

    if hasattr(warning, "__dict__"):
        kwargs = {
            key: getattr(warning, key)
            for key in warning.__dict__
            if not key.startswith("_")
        }

    event("warn", warning=warning, **kwargs)


#--------------------------------------------------
# SourceInfo
#--------------------------------------------------

@dataclass
class SourceInfo:
    file: str = "Unknown"
    line: int = 0
    source: str = ""
    block: ast.AST|None = None

    def modify(self, transformer:ast.NodeTransformer):
        if not self.block:
            raise Exception("Cannot modify source info without a block")

        new_block = transformer.visit(self.block)
        return SourceInfo(self.file, self.line, ast.unparse(new_block), new_block)

    def to_json(self):
        return {
            "file": self.file,
            "line": self.line,
            "source": self.source
        }

#--------------------------------------------------
# Jupyter
#--------------------------------------------------

class Jupyter:
    def __init__(self):
        self.dirty_cells = set()
        try:
            from IPython import get_ipython # type: ignore
            self.ipython = get_ipython()
            if self.ipython:
                self.ipython.events.register('pre_run_cell', self.pre_run_cell)
                self.ipython.events.register('post_run_cell', self.post_run_cell)
                self.dirty_cells.add(self.cell())
        except ImportError:
            self.ipython = None

    def pre_run_cell(self, info):
        self.dirty_cells.add(info.cell_id)

    def post_run_cell(self, result):
        try:
            from . import dsl
            graph = dsl.get_graph()
            if graph._temp_is_active():
                graph._flush_temp()
                graph._restore_temp()
        except Exception:
            return

    def cell_content(self):
        if self.ipython:
            last_input = self.ipython.user_ns['In'][-1]
            return (last_input, f"In[{len(self.ipython.user_ns['In']) - 1}]")
        return ("", "")

    def is_colab(self):
        if not self.ipython:
            return False
        try:
            parent = self.ipython.get_parent() #type: ignore
        except Exception:
            return False
        if not parent or "metadata" not in parent:
            return False
        return "colab" in parent["metadata"]

    def cell(self):
        if self.ipython:
            if self.is_colab():
                return self.ipython.get_parent()["metadata"]["colab"]["cell_id"] #type: ignore
            else:
                try:
                    return self.ipython.get_parent()["metadata"]["cellId"] #type: ignore
                except Exception:
                    return None
        return None

jupyter = Jupyter()

#--------------------------------------------------
# Position capture
#--------------------------------------------------

rai_site_packages = os.path.join("site-packages", "relationalai")
rai_src = os.path.join("src", "relationalai")

def first_non_relationalai_frame(frame):
    while frame and frame.f_back:
        file = frame.f_code.co_filename
        if rai_site_packages not in file and rai_src not in file:
            break
        frame = frame.f_back
    return frame

def capture_code_info(steps=None):
    # Get the current frame and go back to the caller's frame
    caller_frame = inspect.currentframe()
    if steps is not None:
        for _ in range(steps):
            if not caller_frame or not caller_frame.f_back:
                break
            caller_frame = caller_frame.f_back
    else:
        caller_frame = first_non_relationalai_frame(caller_frame)

    if not caller_frame:
        return

    caller_filename = caller_frame.f_code.co_filename
    caller_line = caller_frame.f_lineno

    relative_filename = os.path.relpath(caller_filename, os.getcwd())

    # Read the source code from the caller's file
    source_code = None
    try:
        with open(caller_filename, "r") as f:
            source_code = f.read()
    except IOError:
        (jupyter_code, jupyter_cell) = jupyter.cell_content()
        if jupyter_code:
            source_code = jupyter_code
            relative_filename = jupyter_cell

    if not source_code:
        return SourceInfo(relative_filename, caller_line)
    else:
        return find_block_in(source_code, caller_line, relative_filename)

def find_block_in(source_code: str, caller_line: int, relative_filename: str):
    # Parse the source code into an AST
    tree = ast.parse(source_code)

    # Find the node that corresponds to the call
    class BlockFinder(ast.NodeVisitor):
        def __init__(self, target_lineno):
            self.target_lineno = target_lineno
            self.block_node = None

        def generic_visit(self, node):
            if hasattr(node, "lineno") and node.lineno == self.target_lineno:
                self.block_node = node
                # Stop visiting once the target node is found
                return
            ast.NodeVisitor.generic_visit(self, node)

    finder = BlockFinder(caller_line)
    finder.visit(tree)

    if finder.block_node:
        # Extract the lines from the source code
        start_line = finder.block_node.lineno
        end_line = getattr(finder.block_node, "end_lineno", start_line)

        block_lines = source_code.splitlines()[start_line - 1:end_line]
        block_code = "\n".join(block_lines)
        return SourceInfo(relative_filename, caller_line, dedent(block_code), finder.block_node)

    lines = source_code.splitlines()
    if caller_line > len(lines):
        return SourceInfo(relative_filename, caller_line)
    return SourceInfo(relative_filename, caller_line, lines[caller_line - 1])

def check_errors(task:Task|Action):
    class ErrorFinder(ast.NodeVisitor):
        def __init__(self, start_line):
            self.errors = []
            self.start_line = start_line

        def to_line_numbers(self, node):
            return (node.lineno, node.end_lineno)

        def generic_visit(self, node):
            if isinstance(node, ast.If):
                InvalidIfWarning(task, *self.to_line_numbers(node))
            elif isinstance(node, ast.For) or isinstance(node, ast.While):
                InvalidLoopWarning(task, *self.to_line_numbers(node))
            elif isinstance(node, ast.Try):
                InvalidTryWarning(task, *self.to_line_numbers(node))

            ast.NodeVisitor.generic_visit(self, node)

    source = get_source(task)
    if not source or not source.block:
        return
    ErrorFinder(source.line).visit(source.block)


sources:Dict[Task|Action, SourceInfo|None] = {}
def set_source(item, steps=1, dynamic=False):
    if not DEBUG:
        return
    found = capture_code_info()
    if found:
        sources[item] = found
        if not dynamic:
            check_errors(item)
            pass
    return found

def get_source(item):
    return sources.get(item)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

import argparse
import os
import re
from typing import Any, Callable
import argcomplete
from gentest.cli.repro import encode, reproduce, export_ir, export_pyrel
from gentest.cli.run_tests import run_tests
from gentest.cli.watch import run as watch

subparsers: dict[argparse.ArgumentParser, argparse._SubParsersAction] = {}

def add_command(name: str, help: str, args: dict[str, dict], parent:argparse.ArgumentParser|None=None, action: Callable[..., Any]|None = None, allow_unknown = False):

    if parent:
        if parent not in subparsers:
            subparsers[parent] = parent.add_subparsers(dest='subcommand', help=f'Sub-commands for {parent.prog}')
        subparser = subparsers[parent]
        parser = subparser.add_parser(name, help=help)
    else:
        parser = argparse.ArgumentParser(name, description=help)

    for arg_name, arg_info in args.items():
        parser.add_argument(arg_name, **arg_info)

    # Store the action directly in the parser
    if action:
        parser.set_defaults(func=action)

    if allow_unknown:
        parser.set_defaults(allow_unknown=allow_unknown)

    return parser

def to_short_path(value) -> bool:
    if not re.match(r'^[a-zA-Z0-9]+/[a-zA-Z0-9]+$', value):
        raise argparse.ArgumentTypeError(f"'{value}' is not in the format <key_hash>/<value_hash>")
    return value

def to_file_path(path):
    if not os.path.normpath(path) == path:
        raise argparse.ArgumentTypeError(f"'{path}' contains invalid characters for a file path")

    return path



def main():
    parser = add_command('gentest', 'Development tool for generative test suite.', {})

    short_path = {'help': 'Path to the failing example.', 'type': to_short_path}
    out = {'help': 'Output file path to write to.', 'type': to_file_path}
    ix = {'help': 'Which argument of the test should be emitted', 'type': int, 'default': 0}

    add_command('test', 'Run the test suite, passing args through to pytest.', {}, parent=parser, action=run_tests, allow_unknown=True)
    add_command('watch', 'Run the development tool, validating changes as they get made.', {}, parent=parser, action=watch)
    add_command('encode-failure', 'Encode a failure as a base64 string.', {'short_path': short_path}, parent=parser, action=encode)
    add_command('reproduce', 'Reproduce a failing test case.', {'short_path': short_path}, parent=parser, action=reproduce)

    export_parser = add_command('export', 'Export utilities for debugging.', {}, parent=parser)
    add_command('ir', 'Export the intermediate representation for debugging.', {'short_path': short_path, 'out': out}, parent=export_parser, action=export_ir)
    add_command('pyrel', 'Export the Python DSL code for debugging.', {'short_path': short_path, 'out': out, '--ix': ix}, parent=export_parser, action=export_pyrel)

    argcomplete.autocomplete(parser)

    # args = parser.parse_args()
    args, unknown_args = parser.parse_known_args()
    if hasattr(args, 'func'):
        kwargs = {**vars(args)}
        kwargs.pop('func')
        kwargs.pop('allow_unknown', None)
        kwargs.pop('subcommand', None)
        if getattr(args, 'allow_unknown', False):
            kwargs['unknown_args'] = unknown_args
        elif len(unknown_args):
            raise argparse.ArgumentError(None, f"Specified subcommand doesn't consume the following flags: {', '.join(unknown_args)}")
        args.func(**kwargs)
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



import numbers

from .. import dsl

# Custom number type
_Number = numbers.Number | dsl.Producer

# NOTE: Right now, common contains all Rel stdlib relations.
# If the stdlib is split into multiple namespaces, this will have to be updated.
_math_ns = dsl.global_ns.std.common


# ------------------------------
# Basics
# ------------------------------

def abs(value: _Number) -> dsl.Expression:
    return _math_ns.abs(value)

def isclose(x: _Number, y: _Number, tolerance: _Number = 1e-9) -> dsl.Expression:
    return _math_ns.approx_eq(tolerance, x, y)

def cbrt(value: _Number) -> dsl.Expression:
    return _math_ns.cbrt(value)

def log(x: _Number, base: _Number | None = None) -> dsl.Expression:
    if isinstance(x, numbers.Number) and x <= 0:
        raise ValueError("Cannot take the logarithm of a negative number")
    if base is None:
        return _math_ns.natural_log(x)
    return _math_ns.log(base, x)

def sign(x: _Number) -> dsl.Expression:
    return _math_ns.sign(x)

def sqrt(value: _Number) -> dsl.Expression:
    if isinstance(value, numbers.Number) and value < 0:
        raise ValueError("Cannot take the square root of a negative number")
    return _math_ns.sqrt(value)

def trunc_divide(numerator: _Number, denominator: _Number) -> dsl.Expression:
    return _math_ns.trunc_divide(numerator, denominator)


# ------------------------------
# Trigonometry
# ------------------------------

def acos(value: _Number) -> dsl.Expression:
    return _math_ns.acos(value)

def asin(value: _Number) -> dsl.Expression:
    return _math_ns.asin(value)

def atan(value: _Number) -> dsl.Expression:
    return _math_ns.atan(value)

def cos(value: _Number) -> dsl.Expression:
    return _math_ns.cos(value)

def degrees(radians: _Number) -> dsl.Expression:
    return _math_ns.rad2deg(radians)

def radians(degrees: _Number) -> dsl.Expression:
    return _math_ns.deg2rad(degrees)

def sin(value: _Number) -> dsl.Expression:
    return _math_ns.sin(value)

def tan(value: _Number) -> dsl.Expression:
    return _math_ns.tan(value)


# ------------------------------
# Rounding
# ------------------------------

def ceil(value: _Number) -> dsl.Expression:
    return _math_ns.ceil(value)

def floor(value: _Number) -> dsl.Expression:
    return _math_ns.floor(value)


# ------------------------------
# Exports
# ------------------------------

__all__ = [
    "abs",
    "acos",
    "asin",
    "atan",
    "isclose",
    "cbrt",
    "cos",
    "degrees",
    "log",
    "radians",
    "sign",
    "sin",
    "sqrt",
    "tan",
    "trunc_divide",
    "ceil",
    "floor",
]



import relationalai as rai

from relationalai.std import aggregates
from relationalai.clients.snowflake import Snowflake

# derived from https://relationalai.atlassian.net/browse/RAI-22876

span = globals()["span"]
use_multi_valued = True
use_value_types = True

# This benchmark expects the model to already be present, with the data loaded into it.
model = rai.Model("HeapSnapshotsBenchmark", config=globals()["config"])
# Set config flags on model and span
model._config.set("compiler.use_multi_valued", use_multi_valued)
model._config.set("compiler.use_value_types", use_value_types)
span["use_multi_valued"] = use_multi_valued
span["use_value_types"] = use_value_types

sf = Snowflake(model)

Node = model.Type("Node")
ArrayEdge = model.Type("ArrayEdge")
ObjectEdge = model.Type("ObjectEdge")
Edge = model.Type("Edge")

ArrayTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__ARRAYS
ObjectTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__OBJECTS
NodeTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__NODES
StringsTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__STRINGS

with model.query(tag="select_5") as select:
    out = select(5)
print(out.results)

with model.query(tag="count_strings") as select:
    s = StringsTableEntry()
    out = select(aggregates.count(s))
print(out.results)

with model.query(tag="count_string_values") as select:
    s = StringsTableEntry()
    out = select(aggregates.count(s.value))
print(out.results)

# Build the nodes using the string table to get string typename values
with model.rule():
    n = NodeTableEntry()
    name = StringsTableEntry(id = n.type)
    Node.add(id=n.id, typename=name.value)

# Build the edges using the string table to get string typename values
with model.rule():
    o = ObjectTableEntry()
    label = StringsTableEntry(id = o.label)
    ObjectEdge.add(from_=o.from_, to=o.to_, label=label.value)
    Edge.add(from_=o.from_, to=o.to_)

with model.rule():
    a = ArrayTableEntry()
    ArrayEdge.add(from_=a.from_, to=a.to_, index=a.index)
    Edge.add(from_=a.from_, to=a.to_)

with model.query(tag="count_object_edges") as select:
    out = select(aggregates.count(ObjectEdge()))
print(out.results)



from __future__ import annotations
import atexit
import re
from collections import defaultdict
from typing import Dict, List, Any, Tuple, cast, Callable

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pandas import DataFrame
import time

from .types import AvailableModel, Import, ImportSource
from ..clients.config import Config
from ..compiler import Compiler
from .. import dsl, debugging, metamodel as m

#--------------------------------------------------
# InstallBatch
#--------------------------------------------------

class InstallBatch:
    def __init__(self):
        self.dirty = set()
        self.content:Dict[str, Dict[str, List[str]]] = defaultdict(lambda: defaultdict(list))
        self.control_items:List[Tuple[str, Callable|None]] = []
        self.task_map = defaultdict(dict)

    def _cell(self):
        return debugging.jupyter.cell() or ""

    def _check_dirty_cells(self, name:str):
        for cell in debugging.jupyter.dirty_cells:
            self.content[name][cell].clear()
        debugging.jupyter.dirty_cells.clear()

    def append(self, name:str, code:str, task:m.Task|None=None):
        self._check_dirty_cells(name)
        self.dirty.add(name)
        self.content[name][self._cell()].append((code, task))

    def set(self, name:str, code:str, task:m.Task|None=None):
        self.dirty.add(name)
        self.content[name][self._cell()] = [(code, task)]

    def flush(self):
        items = []
        for name in self.dirty:
            all_cells = []
            cur_line = 0
            task_map = self.task_map[name]
            for _, content in self.content[name].items():
                for (code, task) in content:
                    all_cells.append(code)
                    end = code.count("\n") + cur_line + 2
                    task_map[task] = (cur_line, end)
                    cur_line = end
            items.append((name, "\n\n".join(all_cells)))
        self.dirty.clear()
        return items

    def get_all_models(self):
        items = []
        for name in self.content:
            all_cells = []
            for _, content in self.content[name].items():
                for code, _ in content:
                    all_cells.append(code)
            items.append((name, "\n\n".join(all_cells)))
        return items

    def flush_control_items(self):
        cur = self.control_items
        self.control_items = []
        return cur

    def is_dirty(self):
        return len(self.dirty) > 0

    def line_to_task(self, name:str, line:int):
        for task, (start, end) in self.task_map[name].items():
            if start <= line <= end:
                return task
        return None

#--------------------------------------------------
# Helpers
#--------------------------------------------------
@dataclass
class ExportParams:
    source_database: str
    database: str
    engine: str
    func_name: str
    inputs: List[Tuple[str, str, Any]]
    out_fields: List[Tuple[str, Any]]
    code: str
    dry_run: bool

#--------------------------------------------------
# Resources
#--------------------------------------------------

class ResourceProvider(ABC):
    def __init__(self, profile: str | None = None, config:Config|None=None):
        super().__init__()
        self.config = config or Config(profile)
        self.delete_on_exit = True
        atexit.register(self.cancel_pending_transactions)

    @property
    def platform(self):
        return self.config.get("platform")

    @abstractmethod
    def reset(self):
        pass

    #--------------------------------------------------
    # Generic
    #--------------------------------------------------

    @abstractmethod
    def get_version(self):
        pass

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------

    @abstractmethod
    def get_engine_sizes(self) -> List[Any]:
        pass

    @abstractmethod
    def list_engines(self, state: str|None = None) -> List[Any]:
        pass

    @abstractmethod
    def get_engine(self, name: str):
        pass

    @abstractmethod
    def is_valid_engine_state(self, name: str):
        pass

    @abstractmethod
    def create_engine(self, name: str, size: str, pool:str=""):
        pass

    @abstractmethod
    def delete_engine(self, name: str):
        pass

    @abstractmethod
    def suspend_engine(self, name: str):
        pass

    @abstractmethod
    def resume_engine(self, name: str):
        pass

    def get_default_engine_name(self) -> str:
        engine = self.config.get("engine")
        if not engine:
            raise Exception("No default engine in your configuration")
        return engine

    def get_app_name(self) -> str:
        return cast(str, self.config.get("rai_app_name", "relationalai"))

    #--------------------------------------------------
    # Transactions
    #--------------------------------------------------
    @abstractmethod
    def get_transaction(self, transaction_id):
        pass

    @abstractmethod
    def list_transactions(self, limit:int, only_active=False, **kwargs) -> List[dict]:
        pass

    @abstractmethod
    def cancel_transaction(self, transaction_id):
        pass

    @abstractmethod
    def cancel_pending_transactions(self):
        pass
    
    @abstractmethod
    def get_transaction_events(self, transaction_id:str, continuation_token:str) -> dict:
        pass

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    @abstractmethod
    def list_graphs(self) -> List[AvailableModel]:
        pass

    @abstractmethod
    def get_graph(self, name:str):
        pass

    @abstractmethod
    def create_graph(self, name: str):
        pass

    @abstractmethod
    def delete_graph(self, name: str):
        pass

    @abstractmethod
    def clone_graph(self, target_name: str, source_name: str, nowait_durable: bool = True):
        pass

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    @abstractmethod
    def get_database(self, database: str):
        pass

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    @abstractmethod
    def list_models(self, database: str, engine: str):
        pass

    @abstractmethod
    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        pass

    @abstractmethod
    def delete_model(self, database: str, engine: str, name: str):
        pass

    @abstractmethod
    def create_models_code(self, models:List[Tuple[str, str]]) -> str:
        pass

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    @abstractmethod
    def list_exports(self, database: str, engine: str):
        pass

    @abstractmethod
    def create_export(self, params: ExportParams):
        pass

    @abstractmethod
    def create_export_table(self, database: str, engine: str, table: str, relation: str, columns: Dict[str, str], code: str, refresh: str|None=None):
        pass

    @abstractmethod
    def delete_export(self, database: str, engine: str, name: str):
        pass

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    @abstractmethod
    def list_imports(self, id:str|None, name:str|None, model:str|None, status:str|None, creator:str|None) -> list[Import]:
        pass

    @abstractmethod
    def poll_imports(self, sources:List[str], model:str):
        pass

    @abstractmethod
    def get_import_stream(self, id:str|None, name:str|None):
        pass

    @abstractmethod
    def create_import_stream(self, source: ImportSource, model: str, rate: int, resume: str|None, suspend: str|None, options: dict|None):
        pass

    @abstractmethod
    def change_stream_status(self, stream_id: str, model: str, suspend: bool):
        pass

    @abstractmethod
    def set_imports_engine(self, engine: str):
        pass

    @abstractmethod
    def change_imports_status(self, suspend: bool):
        pass

    @abstractmethod
    def create_import_snapshot(self, source: ImportSource, model: str, options: dict|None):
        pass

    @abstractmethod
    def delete_import(self, import_name: str, model: str):
        pass

    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    @abstractmethod
    def exec_raw(self, database: str, engine: str, raw_code: str, readonly: bool = True, inputs={}, nowait_durable=False, headers={}) -> Any: # @FIXME: Better type annotation
        pass

    def exec_format(self, database: str, engine: str, raw_code: str, task:m.Task, format:str, readonly: bool = True, inputs={}) -> Any: # @FIXME: Better type annotation
        pass

    @abstractmethod
    def format_results(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        pass

    #--------------------------------------------------
    # Types
    #--------------------------------------------------

    @abstractmethod
    def to_model_type(self, model:dsl.Graph, name:str, source:str) -> dsl.Type:
        pass


#--------------------------------------------------
# Client
#--------------------------------------------------

class Client():
    def __init__(
        self,
        resources: ResourceProvider,
        compiler: Compiler,
        database: str,
        config: Config,
        dry_run=False,
        isolated=True,
        keep_model=False,
        nowait_durable=True,
    ):
        self.dry_run = dry_run
        self._source_database = database
        self._database = database[:30] + "_" + config.get_hash() if isolated else database
        self.compiler = compiler
        self._install_batch = InstallBatch()
        self.resources = resources
        self.keep_model = keep_model
        self.isolated = isolated
        self.batch_span: debugging.Span|None = None

        if not self.dry_run:
            self.create_database(isolated=isolated, nowait_durable=nowait_durable)

    def create_database(self, isolated=True, nowait_durable=True):
        start = time.perf_counter()
        with debugging.span("create_database", source=self._source_database, target=self._database):
            database_exists = self.resources.get_graph(self._source_database)
            if isolated:
                if not database_exists:
                    self.resources.create_graph(self._source_database)
                temp = self._database
                self._database = self._source_database
                self.exec_raw("", readonly=False, internal=True, nowait_durable=True)
                self._database = temp
                self.resources.clone_graph(self._database, self._source_database, nowait_durable=nowait_durable)
            elif not database_exists:
                self.resources.create_graph(self._database)
            if isolated and not self.keep_model:
                atexit.register(self.delete_database)
            debugging.time("create_database", time.perf_counter() - start)

    def delete_database(self):
        if self.resources.delete_on_exit and not self.dry_run:
            self.resources.delete_graph(self._database)

    def get_engine_name(self, name:str|None=None) -> str:
        return str(name or self.resources.config.get("engine"))

    def report_errors(self, errors:List[Any], abort_on_error=True, task=None):
        maybe_abort = False
        undefineds = []
        if len(errors):
            for problem in errors:
                if problem.get("is_error") or problem.get("is_exception"):
                    maybe_abort = True
                    message = problem.get("message", "")
                    report = problem.get("report", "")
                    path = problem.get("path", "")
                    if problem.get("error_code") == "UNDEFINED_IDENTIFIER":
                        match = re.search(r'`(.+?)` is undefined', message)
                        line = report.split("|")[0]
                        if line:
                            line = int(line) - 1
                        source_task = self._install_batch.line_to_task(path, line) or task
                        source = debugging.get_source(source_task)
                        if not source:
                            source = debugging.SourceInfo()
                        undefineds.append((match.group(1), source))
                    else:
                        if message:
                            print(message)
                        if report:
                            print(report)
        if abort_on_error and maybe_abort:
            from relationalai.errors import RelQueryError, UninitializedPropertyException
            if undefineds:
                raise UninitializedPropertyException(undefineds)
            raise RelQueryError(errors)

    def load_raw_file(self, path:str):
        content = open(path).read()
        code = self.compiler.compile(dsl.build.raw_task(content))
        self._install_batch.set(path, code)

    def exec_raw(self, code:str, readonly=True, raw_results=True, inputs={}, internal=False, nowait_durable=None, headers={}):
        return self.query(dsl.build.raw_task(code), readonly=readonly, raw_results=raw_results, inputs=inputs, internal=internal, nowait_durable=nowait_durable, headers=headers)

    def exec_control(self, code:str, cb:Callable|None=None):
        self._install_batch.control_items.append((code, cb))

    def install_raw(self, code:str, name:str="pyrel_batch_0"):
        if not name:
            name = "pyrel_batch_0"
        self.compiler.compile(dsl.build.raw_task(code))
        self._install_batch.append(name, code)

    def _query(self, code:str, task:m.Task|None, end_span, readonly=True, raw_results=False, inputs={}, nowait_durable=None, headers={}):
        if nowait_durable is None:
            nowait_durable = self.isolated

        results = self.resources.exec_raw(self._database, self.get_engine_name(), code, readonly=readonly, inputs=inputs, nowait_durable=nowait_durable, headers=headers)
        dataframe, errors = self.resources.format_results(results, task)
        end_span["results"] = dataframe
        end_span["errors"] = errors
        self.report_errors(errors, task=task, abort_on_error=not raw_results)
        return results if raw_results else dataframe

    def _query_format(self, code:str, task:m.Task, end_span, format, readonly=True, inputs={}):
        results, errors = self.resources.exec_format(self._database, self.get_engine_name(), code, task, readonly=readonly, inputs=inputs, format=format)
        # dataframe, errors = self.resources.format_results(results, task)
        end_span["results"] = results
        end_span["errors"] = errors
        # self.report_errors(errors, task=task, abort_on_error=not raw_results)
        # return results if raw_results else dataframe
        return results

    def query(self, task:m.Task, rentrant=False, readonly=True, raw_results=False, inputs={}, format="pandas", tag=None, nowait_durable=None, headers={}, internal=False) -> DataFrame|Any:
        self._install_batch_flush()

        with debugging.span("query", model=self._database, task=task, tag=tag, internal=internal) as end_span:
            code = self.compiler.compile(task)
            if task.has_persist():
                readonly = False
            if self.dry_run:
                return DataFrame()

            start = time.perf_counter()
            if format == "pandas":
                results = self._query(code, task, end_span, readonly=readonly, raw_results=raw_results, inputs=inputs, nowait_durable=nowait_durable, headers=headers)
                debugging.time("query", time.perf_counter() - start, DataFrame() if raw_results else results, internal=internal)
            else:
                results = self._query_format(code, task, end_span, readonly=readonly, inputs=inputs, format=format)
                debugging.time("query", time.perf_counter() - start, DataFrame())
            return results

    def _install_batch_flush(self):
        if not self._install_batch.is_dirty():
            return

        debugging.span_end(self.batch_span)
        self.batch_span = None

        if not self.dry_run:
            with debugging.span("install_batch", model=self._database) as end_span:
                start = time.perf_counter()
                rules = self._install_batch.flush()
                control_items = self._install_batch.flush_control_items()
                control_code = "\n\n".join([c[0] for c in control_items])
                rule_code = self.resources.create_models_code(rules)
                install_code = control_code + rule_code
                dataframe = self._query(install_code, None, end_span, readonly=False, raw_results=True)
                for (_, cb) in control_items:
                    if cb:
                        cb(dataframe)
                debugging.time("install_batch", time.perf_counter() - start, code=install_code)

    def get_install_models(self):
        return self._install_batch.get_all_models()

    def install(self, name, task:m.Task):
        if not self.batch_span:
            self.batch_span = debugging.span_start("rule_batch")
        with debugging.span("rule", model=self._database, task=task, name=name):
            code = self.compiler.compile(task)
            self._install_batch.append("pyrel_batch_0", code, task=task)

    def export_udf(self, name, inputs:List[Tuple[str, m.Var, Any]], outputs, task:m.Task, engine=""):
        self._install_batch_flush()

        cols = task.return_cols()
        if len(outputs) != len(cols):
            raise Exception(f"Expected {len(outputs)} outputs, got {len(cols)}")
        rel_code = self.compiler.compile(task)
        emitted_inputs = [(name, self.compiler.emitter.emit(var), type) for (name, var, type) in inputs]
        outputs = list(zip(cols, outputs))
        if not engine:
            engine = self.get_engine_name()
        params = ExportParams(
            source_database=self._source_database,
            database=self._database,
            engine=engine,
            func_name=name,
            inputs=emitted_inputs,
            out_fields=outputs,
            code=rel_code,
            dry_run=self.dry_run
        )
        self.resources.create_export(params)

    def export_table(self, relation, name, cols, task:m.Task, engine="", refresh=""):
        code = self.compiler.compile(task)
        if not engine:
            engine = self.get_engine_name()
        if not self.dry_run:
            self.resources.create_export_table(self._database, engine, name, relation, cols, code, refresh=refresh)




import pytest

from pathlib import Path
from relationalai.clients.config import ConfigStore, items_with_dots

@pytest.fixture
def basic_config_store():
    return ConfigStore(toml_string=get_config_test_file("raiconfig_example.toml"))

def get_config_test_file(name: str):
    with open(Path(__file__).parent / f"test_configs/{name}", "r") as f:
        return f.read()

def test_config(basic_config_store):
    config = basic_config_store.get_config()
    assert str(config) == get_config_test_file("printed_config.txt")

def test_new_profile(basic_config_store):
    config = basic_config_store.get_config()
    config.set("compiler.dry_run", True)
    config.profile = "foo"
    computed = basic_config_store.with_new_profile(config).as_string()
    actual = get_config_test_file("config_with_new_profile.toml")
    assert computed == actual

def test_items_with_dots():
    D = {
        "a": {
            "b": {
                "c": 1
            }
        },
        "d": 2
    }
    assert tuple(items_with_dots(D)) == (("a.b.c", 1), ("d", 2))




# pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_effects"), config=globals().get("config"))

Person = model.Type("Person")
Adult = model.Type("Adult")
Woop = model.Type("Woop")
Zomg = model.Type("Zomg")

with model.rule():
    joe = Person.add(name="Joe", age=74, key=10, source=100, description="Joe is a cool guy", subbrand=1)
    bob = Person.add(name="Bob", age=40, key=20, source=200, description="Bob is a cool guy", subbrand=2)
    jane = Person.add(name="Jane", age=10, key=30, source=300, description="Jane is a cool gal", subbrand=3)

with model.rule():
    raw = Person()
    i = Adult.add(id=raw.key, source = raw.source).set(Zomg)
    Woop.add(id = raw.subbrand, type= 'raw2').set(Zomg)
    Woop.add(id = raw.subbrand, type= 'raw3').set(Zomg)
    Woop.add(id = raw.subbrand, type= 'raw4').set(Zomg)
    Woop.add(id = raw.subbrand, type= 'raw5').set(Zomg)

with model.query() as select:
    w = Woop()
    z = select(w, w.id, w.type)


from typing import cast
from hypothesis import strategies as gen
from hypothesis.strategies import SearchStrategy as Strategy
from gentest.gen.group_limited import group_limiter
from gentest.gen.scope import GenScope
from gentest.gen.staged import staged
from gentest.gen import ir
from gentest.gen.context import GenContext

#-------------------------------------------------------------------------------
# Provisioning
#-------------------------------------------------------------------------------

def gen_entity_provider(ctx: GenContext) -> Strategy[ir.EntityProvider]:
    return ctx.sample_non_primitive_type().flatmap(
        lambda type: gen.builds(
            ir.EntityProvider,
            entity = gen.builds(ir.Var, type=gen.just(type), name=gen.just(f"{type.name}")),
            properties = gen.lists(gen.sampled_from(type.properties), unique=True) if type.properties else gen.just([])))

def gen_computed_provider(scope: GenScope) -> Strategy[ir.ComputedProvider]:
    # @FIXME: This doesn't work if the scope has no vars in it...
    if not scope.vars:
        return gen.nothing()
    computed_strat = scope.gen_computed(None)
    agg_strat = scope.gen_agg(None)
    return gen.builds(
        ir.ComputedProvider,
        group_limiter(
            gen.one_of(
                computed_strat if computed_strat is not None else  gen.nothing(),
                agg_strat if agg_strat is not None else  gen.nothing()),
            expr=6))

def gen_provider(scope: GenScope) -> Strategy[ir.Provider]:
    return gen.one_of(
        gen_computed_provider(scope),
        gen_entity_provider(scope.context))

def gen_provision_stage(root: GenScope, min_size=0, max_size=5) -> Strategy[tuple[list[ir.Action], GenScope]]:
    def populate_scope(providers: list[ir.Provider]) -> tuple[list[ir.Action], GenScope]:
        sub_builder = root.build()
        for item in providers:
            sub_builder.provide(item)

        return (cast(list[ir.Action], providers), sub_builder.finish())
    return gen.lists(gen_provider(root), min_size=min_size, max_size=max_size).map(populate_scope)

#-------------------------------------------------------------------------------
# Consumption
#-------------------------------------------------------------------------------

def gen_consumer(scope: GenScope):
    return gen.one_of(
        scope.gen_filter(),
        scope.gen_assert())

def gen_consumption_stage(scope: GenScope, *, min_size=0, max_size=15) -> Strategy[tuple[list[ir.Action], GenScope]]:
    return gen.lists(gen_consumer(scope), min_size=min_size, max_size=max_size).map(lambda body: (cast(list[ir.Action], body), scope))

#-------------------------------------------------------------------------------
# Staged Generation
#-------------------------------------------------------------------------------

def gen_stage(actions: list[ir.Action], scope: GenScope):
    return (gen_consumption_stage(scope) | gen_provision_stage(scope)).map(lambda pair: (actions + pair[0], pair[1]))

def gen_task_body(root: GenScope, max_stages: int):
    return gen_provision_stage(root, min_size=1, max_size=5).flatmap(lambda pair: staged(gen_stage, pair, lambda actions, _: len(actions) > 50, max_stages=max_stages))



from gentest import fixtures
from gentest.gen.context import fixture
from gentest.gen.scope import GenScope
from gentest.validate.diff import diff_to_str
from hypothesis import settings, given, strategies as gen
from hypothesis.internal.conjecture.data import ConjectureData
from relationalai import metamodel as mm
from relationalai.clients.test import Executor

from gentest.emit import DocumentEmitter
from gentest.validate.mapping import map_task_ids, normalize_expected_task
from gentest.util import Query, fmt_task
from gentest.harness.database import HYPO_DB
from gentest.gen.task import gen_task
from gentest.validate.errors import RoundtripError, EmitError, ExecError, DiffError
from gentest.validate.roundtrip import roundtripper, exec_and_run_callback

# @NOTE: For reasons I don't understand, stack line numbers change between runs when we directly raise here instead of return the errors
#        which triggers the flaky test detection, so this weird pattern of returning them is necessary.
def task_roundtrip(expected: mm.Task, conjecture: ConjectureData) -> RoundtripError|None:
    __tracebackhide__ = True

    normalize_expected_task(expected)

    doc = DocumentEmitter()
    try:
        doc.rule(expected)
        code = doc.stringify().strip()
    except Exception as err:
        # @NOTE: unfortunately including these tracebacks incorrectly triggers Hypothesis's flaky test detection, because
        # it seems to expect the exact same error to be returned from the same position every shrink.
        # This doesn't make any sense to me, but I don't have more time to investigate it right now.
        # tb = sys.exc_info()[2]
        return EmitError(expected, conjecture.buffer, err, fmt_task) # .with_traceback(tb)

    try:
        actual = exec_and_run_callback(code, retrieve_task)
    except Exception as err:
        # tb = sys.exc_info()[2]
        return ExecError(expected, conjecture.buffer, code, err, fmt_task) # .with_traceback(tb)

    map_task_ids(expected, actual)
    expected_str = fmt_task(expected)
    actual_str = fmt_task(actual)
    if expected_str != actual_str:
        return DiffError(expected, conjecture.buffer, code, actual, diff_to_str(expected_str, actual_str))


def retrieve_task(exec: Executor) -> mm.Task:
    for block in exec.blocks:
        if isinstance(block, Query):
            continue
        if block.name == "pyrelstd":
            continue
        return block.task

    raise Exception("Could not find rule")


ctx = fixture(fixtures.person_place_thing).finish()
root = GenScope(ctx)

@roundtripper
# @seed(3) # reproduces flaky test results
# @reproduce_failure("6.92.0", b'AXicY2RiYGBkYGRkQAAAAHIABw==') # Repros two vars ground by same ent, prop
@settings(max_examples=50, database=HYPO_DB, )
@given(gen_task(root), gen.data())
def test_task_roundtrip(task: mm.Task, data: gen.DataObject):
    err = task_roundtrip(task, data.conjecture_data)
    if err:
        raise err

